
// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// HexapodFrame class - hexapod machine frame

#include "HexapodFrame.H"

HexapodFrame::HexapodFrame
  (char* name, const FrameParameters& parameters) : Artifact(name)
{
   // copy frame parameters
   bottomHeight = parameters.bottomHeight;
   centreHeight = parameters.centreHeight;
   topHeight = parameters.topHeight;

   bottomWidth = parameters.bottomWidth;
   centreWidth = parameters.centreWidth;
   topWidth = parameters.topWidth;

   tubeRadius = parameters.tubeRadius;
   sphereRadius = parameters.sphereRadius;

   // derive other values
   bottomDiagonal = bottomWidth*sqrt(2);
   bottomCorner1 = Vector3(-bottomDiagonal/2, 0, bottomHeight);
   bottomCorner2 = Vector3(0, -bottomDiagonal/2, bottomHeight);
   bottomCorner3 = Vector3(bottomDiagonal/2, 0, bottomHeight);
   bottomCorner4 = Vector3(0, bottomDiagonal/2, bottomHeight);

   centreCorner1 = Vector3(-centreWidth/2, -centreWidth/2, centreHeight);
   centreCorner2 = Vector3(centreWidth/2, -centreWidth/2, centreHeight);
   centreCorner3 = Vector3(centreWidth/2, centreWidth/2, centreHeight);
   centreCorner4 = Vector3(-centreWidth/2, centreWidth/2, centreHeight);

   topDiagonal = topWidth*sqrt(2);
   topCorner1 = Vector3(-topDiagonal/2, 0, topHeight);
   topCorner2 = Vector3(0, -topDiagonal/2, topHeight);
   topCorner3 = Vector3(topDiagonal/2, 0, topHeight);
   topCorner4 = Vector3(0, topDiagonal/2, topHeight);

   ball = HexapodFrameBall("ball", sphereRadius);

   // build hexapod frame

   // bottom frame balls

   add(&ball, bottomCorner1);
   add(&ball, bottomCorner2);
   add(&ball, bottomCorner3);
   add(&ball, bottomCorner4);

   // centre frame balls

   add(&ball, centreCorner1);
   add(&ball, centreCorner2);
   add(&ball, centreCorner3);
   add(&ball, centreCorner4);

   // top frame balls

   add(&ball, topCorner1);
   add(&ball, topCorner2);
   add(&ball, topCorner3);
   add(&ball, topCorner4);

   // bottom frame tubes

   tube1 = HexapodFrameTube("tube1", tubeRadius, bottomCorner1, bottomCorner2);
   add(&tube1);

   tube2 = HexapodFrameTube("tube2", tubeRadius, bottomCorner2, bottomCorner3);
   add(&tube2);

   tube3 = HexapodFrameTube("tube3", tubeRadius, bottomCorner3, bottomCorner4);
   add(&tube3);

   tube4 = HexapodFrameTube("tube4",  tubeRadius, bottomCorner4, bottomCorner1);
   add(&tube4);

   // centre frame tubes

   tube5 = HexapodFrameTube("tube5", tubeRadius, centreCorner1, centreCorner2);
   add(&tube5);

   tube6 = HexapodFrameTube("tube6", tubeRadius, centreCorner2, centreCorner3);
   add(&tube6);

   tube7 = HexapodFrameTube("tube7", tubeRadius, centreCorner3, centreCorner4);
   add(&tube7);

   tube8 = HexapodFrameTube("tube8",  tubeRadius, centreCorner4, centreCorner1);
   add(&tube8);

   // top frame tubes

   tube9 = HexapodFrameTube("tube9", tubeRadius, topCorner1, topCorner2);
   add(&tube9);

   tube10 = HexapodFrameTube("tube10", tubeRadius, topCorner2, topCorner3);
   add(&tube10);

   tube11 = HexapodFrameTube("tube11", tubeRadius, topCorner3, topCorner4);
   add(&tube11);

   tube12 = HexapodFrameTube("tube12",  tubeRadius, topCorner4, topCorner1);
   add(&tube12);

   // bottom to centre connections

   tube13 = 
      HexapodFrameTube("tube13",  tubeRadius, bottomCorner1, centreCorner1);
   add(&tube13);

   tube14 = 
      HexapodFrameTube("tube14",  tubeRadius, bottomCorner1, centreCorner4);
   add(&tube14);

   tube15 = 
      HexapodFrameTube("tube15",  tubeRadius, bottomCorner2, centreCorner1);
   add(&tube15);

   tube16 = 
      HexapodFrameTube("tube16",  tubeRadius, bottomCorner2, centreCorner2);
   add(&tube16);

   tube17 = 
      HexapodFrameTube("tube17",  tubeRadius, bottomCorner3, centreCorner2);
   add(&tube17);

   tube18 = 
      HexapodFrameTube("tube18",  tubeRadius, bottomCorner3, centreCorner3);
   add(&tube18);

   tube19 = 
      HexapodFrameTube("tube19",  tubeRadius, bottomCorner4, centreCorner3);
   add(&tube19);

   tube20 = 
      HexapodFrameTube("tube20",  tubeRadius, bottomCorner4, centreCorner4);
   add(&tube20);

   // top to centre connections

   tube21 = HexapodFrameTube("tube21",  tubeRadius, topCorner1, centreCorner1);
   add(&tube21);

   tube22 = HexapodFrameTube("tube22",  tubeRadius, topCorner1, centreCorner4);
   add(&tube22);

   tube23 = HexapodFrameTube("tube23",  tubeRadius, topCorner2, centreCorner1);
   add(&tube23);

   tube24 = HexapodFrameTube("tube24",  tubeRadius, topCorner2, centreCorner2);
   add(&tube24);

   tube25 = HexapodFrameTube("tube25",  tubeRadius, topCorner3, centreCorner2);
   add(&tube25);

   tube26 = HexapodFrameTube("tube26",  tubeRadius, topCorner3, centreCorner3);
   add(&tube26);

   tube27 = HexapodFrameTube("tube27",  tubeRadius, topCorner4, centreCorner3);
   add(&tube27);

   tube28 = HexapodFrameTube("tube28",  tubeRadius, topCorner4, centreCorner4);
   add(&tube28);
}

