
// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// HexapodLeg class

#include "HexapodLeg.H"

HexapodLeg::HexapodLeg
  (char* name, const LegParameters& parameters) : Artifact(name)
{
   // copy parameters
   screwDiameter = parameters.screwDiameter;
   screwLength = parameters.screwLength;
   screwTopLength = parameters.screwTopLength;
   screwBottomLength = parameters.screwBottomLength;
   nutDiameter = parameters.nutDiameter;
   nutLength = parameters.nutLength;
   nutTubeLength = parameters.nutTubeLength;
   nutTubeOD = parameters.nutTubeOD;
   nutTubeID = parameters.nutTubeID;
   nutTubeBottomLength = parameters.nutTubeBottomLength;
   shroudTubeDiameter = parameters.shroudTubeDiameter;
   shroudTubeLength = parameters.shroudTubeLength;
   totalBacklash = parameters.totalBacklash;
   legLength = parameters.legLength;

   minLength = screwTopLength + nutTubeLength + nutTubeBottomLength;
   maxLength = 
      screwTopLength + screwLength - nutLength +
      nutTubeLength + nutTubeBottomLength;

   // construct leg from primitives, with origin at top end, z axis along c/l
   topEnd = Vector3(0, 0, 0);
   bottomEnd = Vector3(0, 0, legLength);


   // top section
   topLength = Cylinder
     (Vector3(0, 0, 0), 
      Vector3(0, 0, screwTopLength), shroudTubeDiameter/2);
   add(&topLength);

   screw = Cylinder
     (Vector3(0, 0, screwTopLength),
      Vector3(0, 0, screwTopLength + screwLength), screwDiameter/2);
   add(&screw);

   // bottom section
   nutTubeBottom = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength),
      Vector3(0, 0, legLength), nutTubeOD/2);
   add(&nutTubeBottom);

   nutTube = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength - nutTubeLength),
      Vector3(0, 0, legLength - nutTubeBottomLength), nutTubeOD/2);
   add(&nutTube);
}

void
HexapodLeg::operator= 
  (const HexapodLeg& that)
{
   // mirror that to this
   memcpy(this, &that, sizeof(*this));

   // fix artifact components
   firstPrimitive = 0;
   add(&topLength);
   add(&screw);
   add(&nutTubeBottom);
   add(&nutTube);
}

void
HexapodLeg::setParameters
  (const LegParameters& parameters)
{
   // copy parameters
   screwDiameter = parameters.screwDiameter;
   screwLength = parameters.screwLength;
   screwTopLength = parameters.screwTopLength;
   screwBottomLength = parameters.screwBottomLength;
   nutDiameter = parameters.nutDiameter;
   nutLength = parameters.nutLength;
   nutTubeLength = parameters.nutTubeLength;
   nutTubeOD = parameters.nutTubeOD;
   nutTubeID = parameters.nutTubeID;
   nutTubeBottomLength = parameters.nutTubeBottomLength;
   shroudTubeDiameter = parameters.shroudTubeDiameter;
   shroudTubeLength = parameters.shroudTubeLength;
   legLength = parameters.legLength;

   minLength = screwTopLength + nutTubeLength + nutTubeBottomLength;
   maxLength = 
      screwTopLength + screwLength - nutLength +
      nutTubeLength + nutTubeBottomLength;

   topEnd = Vector3(0, 0, 0);
   bottomEnd = Vector3(0, 0, legLength);

   // top section
   topLength = Cylinder
     (Vector3(0, 0, 0),
      Vector3(0, 0, screwTopLength), shroudTubeDiameter/2);

   screw = Cylinder
     (Vector3(0, 0, screwTopLength),
      Vector3(0, 0, screwTopLength + screwLength), screwDiameter/2);

   // bottom section
   nutTubeBottom = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength),
      Vector3(0, 0, legLength), nutTubeOD/2);

   nutTube = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength - nutTubeLength),
      Vector3(0, 0, legLength - nutTubeBottomLength), nutTubeOD/2);
}

void
HexapodLeg::setLength
  (double length)
{
   legLength = length;

   // bottom section
   nutTubeBottom = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength),
      Vector3(0, 0, legLength), nutTubeOD/2);

   nutTube = Cylinder
     (Vector3(0, 0, legLength - nutTubeBottomLength - nutTubeLength),
      Vector3(0, 0, legLength - nutTubeBottomLength), nutTubeOD/2);
}

bool 
HexapodLeg::inRange()
{
   return ((legLength >= minLength) && (legLength <= maxLength));
}

bool 
HexapodLeg::tooLong()
{
   return (legLength > maxLength);
}

bool 
HexapodLeg::tooShort()
{
   return (legLength < minLength);
}
