
// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// HexapodMechanism - the main mechanism

#include "HexapodMechanism.H"

HexapodMechanism::HexapodMechanism
 (char* name, const MechanismParameters& mechanismParameters) : Artifact(name)
{
   // copy parameters
   topCircleRadius = mechanismParameters.topCircleRadius;
   bottomCircleRadius = mechanismParameters.bottomCircleRadius;
   legPairSectorAngle = mechanismParameters.legPairSectorAngle;
   topHeight = mechanismParameters.supportHeight;
   bottomHeight = mechanismParameters.platformHeight;
   legParameters = mechanismParameters.legParameters;
   topJointParameters = mechanismParameters.topJointParameters;
   bottomJointParameters = mechanismParameters.bottomJointParameters;
   platformParameters = mechanismParameters.platformParameters;

   // other initialisations
   cutterForce = Vector3(0.0, 0.0, 0.0);
   cutterForcePosition = Vector3(0.0, 0.0, 0.0);   
   forceAC[0] = forceBC[0] = forceAC[1] = 
      forceBC[1] = forceAC[2] = forceBC[2] = 0.0;

   // compute basic coordinates
   baseAngle = legPairSectorAngle/2;

   double xFix = -30; // small x adjustment to make long range symmetrical

   pointA[0] = Vector3
     (topCircleRadius*cos(baseAngle) + xFix, 
      topCircleRadius*sin(baseAngle), topHeight);
   pointB[0] = Vector3
     (topCircleRadius*cos(-baseAngle) + xFix, 
      topCircleRadius*sin(-baseAngle), topHeight);
   pointA[1] = Vector3
     (topCircleRadius*cos(baseAngle + 2*PI/3) + xFix, 
      topCircleRadius*sin(baseAngle + 2*PI/3), topHeight);
   pointB[1] = Vector3
     (topCircleRadius*cos(-baseAngle + 2*PI/3) + xFix, 
      topCircleRadius*sin(-baseAngle + 2*PI/3), topHeight);
   pointA[2] = Vector3
     (topCircleRadius*cos(baseAngle + 4*PI/3) + xFix, 
      topCircleRadius*sin(baseAngle + 4*PI/3), topHeight);
   pointB[2] = Vector3
     (topCircleRadius*cos(-baseAngle + 4*PI/3) + xFix, 
      topCircleRadius*sin(-baseAngle + 4*PI/3), topHeight);

   pointC[0] = Vector3
     (bottomCircleRadius*cos(0) + xFix,
      bottomCircleRadius*sin(0), bottomHeight);
   pointC[1] = Vector3
     (bottomCircleRadius*cos(2*PI/3) + xFix, 
      bottomCircleRadius*sin(2*PI/3), bottomHeight);
   pointC[2] = Vector3
     (bottomCircleRadius*cos(4*PI/3) + xFix, 
      bottomCircleRadius*sin(5*PI/3), bottomHeight);

   /*
   // coordinate printout if needed
   cout << "A[0] = " << pointA[0] << '\n';
   cout << "B[0] = " << pointB[0] << '\n';
   cout << "A[1] = " << pointA[1] << '\n';
   cout << "B[1] = " << pointB[1] << '\n';
   cout << "A[2] = " << pointA[2] << '\n';
   cout << "B[2] = " << pointB[2] << '\n';
   cout << "C[0] = " << pointC[0] << '\n';
   cout << "C[1] = " << pointC[1] << '\n';
   cout << "C[2] = " << pointC[2] << '\n';
   */

   // create legs
   Vector3 end1 = Vector3(0, 0, 0);
   Vector3 end2;

   legParameters.legLength = pointA[0].distanceTo(pointC[0]);
   leg1A = HexapodLeg("leg1A", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg1AInstance = add(&leg1A, end1, pointA[0], end2, pointC[0]);

   legParameters.legLength = pointB[0].distanceTo(pointC[0]);
   leg1B = HexapodLeg("leg1B", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg1BInstance = add(&leg1B, end1, pointB[0], end2, pointC[0]);

   legParameters.legLength = pointA[1].distanceTo(pointC[1]);
   leg2A = HexapodLeg("leg2A", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg2AInstance = add(&leg2A, end1, pointA[1], end2, pointC[1]);

   legParameters.legLength = pointB[1].distanceTo(pointC[1]);
   leg2B = HexapodLeg("leg2B", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg2BInstance = add(&leg2B, end1, pointB[1], end2, pointC[1]);

   legParameters.legLength = pointA[2].distanceTo(pointC[2]);
   leg3A = HexapodLeg("leg3A", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg3AInstance = add(&leg3A, end1, pointA[2], end2, pointC[2]);

   legParameters.legLength = pointB[2].distanceTo(pointC[2]);
   leg3B = HexapodLeg("leg3B", legParameters);
   end2 = Vector3(0, 0, legParameters.legLength);
   leg3BInstance = add(&leg3B, end1, pointB[2], end2, pointC[2]);

   // create platform
   platform = HexapodPlatform("platform", platformParameters);
   platformTranslation = Vector3(0, 0, bottomHeight);
   platformRotation = Vector3(0, 0, 0);
   platformInstance = add(&platform, platformRotation, platformTranslation);

   // define calibration correction names
   correctionName[0]  = "A0X";
   correctionName[1]  = "A0Y";
   correctionName[2]  = "A0Z";
   correctionName[3]  = "A1X";
   correctionName[4]  = "A1Y";
   correctionName[5]  = "A1Z";
   correctionName[6]  = "A2X";
   correctionName[7]  = "A2Y";
   correctionName[8]  = "A2Z";
   correctionName[9]  = "B0X";
   correctionName[10] = "B0Y";
   correctionName[11] = "B0Z";
   correctionName[12] = "B1X";
   correctionName[13] = "B1Y";
   correctionName[14] = "B1Z";
   correctionName[15] = "B2X";
   correctionName[16] = "B2Y";
   correctionName[17] = "B2Z";
   correctionName[18] = "LA0";
   correctionName[19] = "LA1";
   correctionName[20] = "LA2";
   correctionName[21] = "LB0";
   correctionName[22] = "LB1";
   correctionName[23] = "LB2";
   correctionName[24] = "L01";
   correctionName[25] = "L12";
   correctionName[26] = "L20";
}

void
HexapodMechanism::movePlatformTo  // absolute
  (const Vector3 rotation,
   const Vector3 translation)
{
   platformTranslation = translation;
   platformRotation = rotation;
   platformInstance->setTransformation(platformRotation, platformTranslation);

   pointC[0] = Vector3
     (bottomCircleRadius*cos(0),
      bottomCircleRadius*sin(0), 0);
   pointC[1] = Vector3
     (bottomCircleRadius*cos(2*PI/3), 
      bottomCircleRadius*sin(2*PI/3), 0);
   pointC[2] = Vector3
     (bottomCircleRadius*cos(4*PI/3), 
      bottomCircleRadius*sin(5*PI/3), 0);
                    
   pointC[0] = pointC[0].rotate(rotation);
   pointC[1] = pointC[1].rotate(rotation);
   pointC[2] = pointC[2].rotate(rotation);

   pointC[0] += translation;
   pointC[1] += translation;
   pointC[2] += translation;

   Vector3 end1 = Vector3(0, 0, 0);
   Vector3 end2;
   Vector3 rot, trans, local, global;

   double legLength;

   legLength = pointA[0].distanceTo(pointC[0]);
   leg1A.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = pointA[0] - end1;
   local = end2 - end1;
   global = pointC[0] - pointA[0];
   rot = local.rotationsTo(global);
   leg1AInstance->setTransformation(rot, trans);

   legLength = pointB[0].distanceTo(pointC[0]);
   leg1B.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = (pointB[0] - end1);
   local = (end2 - end1);
   global = (pointC[0] - pointB[0]);
   rot = (local.rotationsTo(global));
   leg1BInstance->setTransformation(rot, trans);

   legLength = pointA[1].distanceTo(pointC[1]);
   leg2A.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = (pointA[1] - end1);
   local = (end2 - end1);
   global = (pointC[1] - pointA[1]);
   rot = (local.rotationsTo(global));
   leg2AInstance->setTransformation(rot, trans);

   legLength = pointB[1].distanceTo(pointC[1]);
   leg2B.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = (pointB[1] - end1);
   local = (end2 - end1);
   global = (pointC[1] - pointB[1]);
   rot = (local.rotationsTo(global));
   leg2BInstance->setTransformation(rot, trans);

   legLength = pointA[2].distanceTo(pointC[2]);
   leg3A.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = (pointA[2] - end1);
   local = (end2 - end1);
   global = (pointC[2] - pointA[2]);
   rot = (local.rotationsTo(global));
   leg3AInstance->setTransformation(rot, trans);

   legLength = pointB[2].distanceTo(pointC[2]);
   leg3B.setLength(legLength);

   end2 = Vector3(0, 0, legLength);
   trans = (pointB[2] - end1);
   local = (end2 - end1);
   global = (pointC[2] - pointB[2]);
   rot = (local.rotationsTo(global));
   leg3BInstance->setTransformation(rot, trans);
}

void
HexapodMechanism::movePlatformBy  // relative
  (const Vector3 rotation,
   const Vector3 translation)
{
   platformRotation += rotation;
   platformTranslation += translation;
   movePlatformTo(platformRotation, platformTranslation);
}

void 
HexapodMechanism::applyCutterForce
  (const Vector3 force,
   const Vector3 position)
// register cutter force and location for later processing
{
   cutterForce = force;
   cutterForcePosition = position;
}

double
HexapodMechanism::angleACB
  (Selector selector)
// angle at bottom of leg pair triangle
{
   Vector3 lineAC[3], lineBC[3];
   double angleACB[3];

   for (int n = 0; n < 3; n++)
      {
         // compute leg direction vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
         
         // find angle between vectors by dot product
         angleACB[n] = 
            acos(lineAC[n]*lineBC[n]/(lineAC[n].length()*lineBC[n].length()));
      }

   // return angle
   double angle = 0;  // = 0 avoids spurious warning
   switch (selector)
      {
         case MIN:
            angle = angleACB[0];
            if (angle > angleACB[1])
               angle = angleACB[1];
            if (angle > angleACB[2])
               angle = angleACB[2];
            break;

         case MAX:
            angle = angleACB[0];
            if (angle < angleACB[1])
               angle = angleACB[1];
            if (angle < angleACB[2])
               angle = angleACB[2];
            break;
      }
   return angle;
}

double
HexapodMechanism::angleBAC
  (Selector selector)
// angle at top of leg pair triangle
{
   Vector3 lineAC[3], lineBC[3];
   Vector3 lineAB[3], lineBA[3];
   double angleBAC[3], angleABC[3];

   for (int n = 0; n < 3; n++)
      {
         // compute direction vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
         lineAB[n] = pointB[n] - pointA[n];
         lineBA[n] = pointA[n] - pointB[n];
        
         // find angle between vectors by dot product
         angleBAC[n] = 
            acos(lineAC[n]*lineAB[n]/(lineAC[n].length()*lineAB[n].length()));
         angleABC[n] = 
            acos(lineBC[n]*lineBA[n]/(lineBC[n].length()*lineBA[n].length()));
      }

   // return angle
   double angle = 0;  // = 0 avoids spurious warning 
   switch (selector)
      {
         case MIN:
            angle = angleBAC[0];
            if (angle > angleABC[0])
               angle = angleABC[0];
            if (angle > angleBAC[1])
               angle = angleBAC[1];
            if (angle > angleABC[1])
               angle = angleABC[1];
            if (angle > angleBAC[2])
               angle = angleBAC[2];
            if (angle > angleABC[2])
               angle = angleABC[2];
            break;

         case MAX:
            angle = angleBAC[0];
            if (angle < angleABC[0])
               angle = angleABC[0];
            if (angle < angleBAC[1])
               angle = angleBAC[1];
            if (angle < angleABC[1])
               angle = angleABC[1];
            if (angle < angleBAC[2])
               angle = angleBAC[2];
            if (angle < angleABC[2])
               angle = angleABC[2];
            break;
      }
   return angle;
}

double
HexapodMechanism::angleGCV
  (Selector selector)
// angle between leg pair plane and horizontal
{
   Vector3 lineAC[3];
   Vector3 lineBC[3];
   Vector3 lineCG[3];  // normal to leg pair plane
   Vector3 vertical (0.0, 0.0, -1.0);  // normal to horizontal - sign necessary
   double angleGCV[3];  // angle between normal to leg pair plane and vertical
                        // = angle between leg pair plane and horizontal

   for (int n = 0; n < 3; n++)
      {
         // compute leg pair direction vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
         
         // find leg pair plane normal vectors by cross product
         lineCG[n] = lineAC[n] ^ lineBC[n];

         // find angle between normal vectors by dot product
         angleGCV[n] = acos(lineCG[n]*vertical/lineCG[n].length());
      }

   // return angle
   double angle = 0;  // = 0 avoids spurious warning
   switch (selector)
      {
         case MIN:
            angle = angleGCV[0];
            if (angle > angleGCV[1])
               angle = angleGCV[1];
            if (angle > angleGCV[2])
               angle = angleGCV[2];
            break;
   
         case MAX:
            angle = angleGCV[0];
            if (angle < angleGCV[1])
               angle = angleGCV[1];
            if (angle < angleGCV[2])
               angle = angleGCV[2];
            break;
      }
   return angle;
}

double
HexapodMechanism::angleACF
  (Selector selector)
// External angle between legs and bottom joint long axis. The long axis is 
// the 'fold' axis between a leg pair plane and the platform plane. Angle 
// measured in leg pair plane.

// We want 'external angles' so should subtract PI from angleACF. 
// AngleBCF ought to be right. But look below at what I do at the moment. 

{
   Vector3 lineAC[3], lineBC[3];  // define leg pair planes
   Vector3 lineCG[3];  // normals to leg pair planes 
   Vector3 lineIJ, lineIK;  // define platform plane
   Vector3 lineIL;  // normal to platform plane
   Vector3 lineCF[3];  // intersections of leg pair and platform planes
   double angleACF[3], angleBCF[3];

   for (int n = 0; n < 3; n++)
      {
         // compute leg pair direction vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
         
         // find leg pair plane normal vectors by cross product
         lineCG[n] = lineAC[n] ^ lineBC[n];
      }

   // compute platform direction vectors
   lineIJ = pointC[1] - pointC[0];
   lineIK = pointC[2] - pointC[0];

   // find platform plane normal vector by cross product
   lineIL = lineIJ ^ lineIK;

   for (int n = 0; n < 3; n++)
      {
         // now find fold axis vectors (normal to normals) by cross product
         lineCF[n] = lineCG[n] ^ lineIL;

         // find external angles between leg and axis vectors by dot product
         angleACF[n] = 
            acos(lineAC[n]*lineCF[n]/(lineAC[n].length()*lineCF[n].length()));
         angleBCF[n] = PI - 
            acos(lineBC[n]*lineCF[n]/(lineBC[n].length()*lineCF[n].length()));
      }

   // return angle
   double angle = 0;  // = 0 avoids spurious warning
   switch (selector)
      {
         case MIN:
            angle = angleACF[0];
            if (angle > angleBCF[0])
               angle = angleBCF[0];
            if (angle > angleACF[1])
               angle = angleACF[1];
            if (angle > angleBCF[1])
               angle = angleBCF[1];
            if (angle > angleACF[2])
               angle = angleACF[2];
            if (angle > angleBCF[2])
               angle = angleBCF[2];
            break;
   
         case MAX:
            angle = angleACF[0];
            if (angle < angleBCF[0])
               angle = angleBCF[0];
            if (angle < angleACF[1])
               angle = angleACF[1];
            if (angle < angleBCF[1])
               angle = angleBCF[1];
            if (angle < angleACF[2])
               angle = angleACF[2];
            if (angle < angleBCF[2])
               angle = angleBCF[2];
            break;
      }
   return angle;
}

double
HexapodMechanism::angleGCH
  (Selector selector)
// angle between leg pair planes and platform plane
{
   Vector3 lineAC[3], lineBC[3];  // define leg pair planes
   Vector3 lineCG[3];  // normals to leg pair planes 
   Vector3 lineIJ, lineIK;  // define platform plane
   Vector3 lineIL;  // normal to platform plane
   double angleGCH[3]; // angle between normals to planes = angle between planes
                       // lineCH[n] = lineIL -- slightly untidy

   for (int n = 0; n < 3; n++)
      {
         // compute leg pair direction vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
         
         // find leg pair plane normal vectors by cross product
         lineCG[n] = lineAC[n] ^ lineBC[n];
      }

   // compute platform direction vectors
   lineIJ = pointC[0] - pointC[1];
   lineIK = pointC[0] - pointC[2];

   // find platform plane normal vectors by cross product
   lineIL = lineIJ ^ lineIK;

   for (int n = 0; n < 3; n++)
      // find angle between leg pair plane normal 
      // and  platform plane normal by dot product
      angleGCH[n] = acos(lineCG[n]*lineIL/(lineCG[n].length()*lineIL.length()));

   // return angle
   double angle = 0;  // = 0 avoids spurious warning
   switch (selector)
      {
         case MIN:
            angle = angleGCH[0];
            if (angle > angleGCH[1])
               angle = angleGCH[1];
            if (angle > angleGCH[2])
               angle = angleGCH[2];
            break;
   
         case MAX:
            angle = angleGCH[0];
            if (angle < angleGCH[1])
               angle = angleGCH[1];
            if (angle < angleGCH[2])
               angle = angleGCH[2];
            break;
      }
   return angle;
}

double
HexapodMechanism::legForce
  (Selector selector)
// compute forces developed in mechanism by applied cutter force
{
   Vector3 unitAC[3], unitBC[3];  // unit leg direction vectors 
   Vector3 pointCT[3];  // point translated by force position

   for (int n = 0; n < 3; n++)
      {
         // compute unit leg direction vectors
         unitAC[n] = (pointC[n] - pointA[n]).normalize();
         unitBC[n] = (pointC[n] - pointB[n]).normalize();

         // compute translated bottom positions 
         pointCT[n] = pointC[n] - cutterForcePosition;
      }

/* Equilibrium equations for coefficient matrix and tests below:
   Note that this commented section is not program code!

   // horizontal equilibrium in x, y, z directions:
   cutterForce[VX] = 
      unitAC[0][VX]*forceAC[0] + unitBC[0][VX]*forceBC[0] + 
      unitAC[1][VX]*forceAC[1] + unitBC[1][VX]*forceBC[1] + 
      unitAC[2][VX]*forceAC[2] + unitBC[2][VX]*forceBC[2]

   cutterForce[VY] = 
      unitAC[0][VY]*forceAC[0] + unitBC[0][VY]*forceBC[0] + 
      unitAC[1][VY]*forceAC[1] + unitBC[1][VY]*forceBC[1] + 
      unitAC[2][VY]*forceAC[2] + unitBC[2][VY]*forceBC[2]

   cutterForce[VZ] = 
      unitAC[0][VZ]*forceAC[0] + unitBC[0][VZ]*forceBC[0] + 
      unitAC[1][VZ]*forceAC[1] + unitBC[1][VZ]*forceBC[1] + 
      unitAC[2][VZ]*forceAC[2] + unitBC[2][VZ]*forceBC[2]

   // rotational equilibrium about transformed x axis:
   0 =      // no applied moment
     (unitAC[0][VY]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VY])*forceAC[0] +
     (unitBC[0][VY]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VY])*forceBC[0] +
     (unitAC[1][VY]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VY])*forceAC[1] +
     (unitBC[1][VY]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VY])*forceBC[1] +
     (unitAC[2][VY]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VY])*forceAC[2] +
     (unitBC[2][VY]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VY])*forceBC[2] 

   // rotational equilibrium about transformed y axis:
   0 =      // no applied moment
     (unitAC[0][VX]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VX])*forceAC[0] +
     (unitBC[0][VX]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VX])*forceBC[0] +
     (unitAC[1][VX]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VX])*forceAC[1] +
     (unitBC[1][VX]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VX])*forceBC[1] +
     (unitAC[2][VX]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VX])*forceAC[2] +
     (unitBC[2][VX]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VX])*forceBC[2] 

   // rotational equilibrium about transformed z axis:
   0 =      // no applied moment
     (unitAC[0][VX]*pointCT[0][VY] - unitAC[0][VY]*pointCT[0][VX])*forceAC[0] +
     (unitBC[0][VX]*pointCT[0][VY] - unitBC[0][VY]*pointCT[0][VX])*forceBC[0] +
     (unitAC[1][VX]*pointCT[1][VY] - unitAC[1][VY]*pointCT[1][VX])*forceAC[1] +
     (unitBC[1][VX]*pointCT[1][VY] - unitBC[1][VY]*pointCT[1][VX])*forceBC[1] +
     (unitAC[2][VX]*pointCT[2][VY] - unitAC[2][VY]*pointCT[2][VX])*forceAC[2] +
     (unitBC[2][VX]*pointCT[2][VY] - unitBC[2][VY]*pointCT[2][VX])*forceBC[2] 

*/

   // build coefficient matrix rows by reference to equations above
   Vector6 row0 
     (unitAC[0][VX], unitBC[0][VX], unitAC[1][VX],
      unitBC[1][VX], unitAC[2][VX], unitBC[2][VX]);

   Vector6 row1 
     (unitAC[0][VY], unitBC[0][VY], unitAC[1][VY],
      unitBC[1][VY], unitAC[2][VY], unitBC[2][VY]);

   Vector6 row2 
     (unitAC[0][VZ], unitBC[0][VZ], unitAC[1][VZ],
      unitBC[1][VZ], unitAC[2][VZ], unitBC[2][VZ]);

   Vector6 row3 
     ((unitAC[0][VY]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VY]),
      (unitBC[0][VY]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VY]),
      (unitAC[1][VY]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VY]),
      (unitBC[1][VY]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VY]),
      (unitAC[2][VY]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VY]),
      (unitBC[2][VY]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VY]));

   Vector6 row4 
     ((unitAC[0][VX]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VX]),
      (unitBC[0][VX]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VX]),
      (unitAC[1][VX]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VX]),
      (unitBC[1][VX]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VX]),
      (unitAC[2][VX]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VX]),
      (unitBC[2][VX]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VX]));

   Vector6 row5 
     ((unitAC[0][VX]*pointCT[0][VY] - unitAC[0][VY]*pointCT[0][VX]),
      (unitBC[0][VX]*pointCT[0][VY] - unitBC[0][VY]*pointCT[0][VX]),
      (unitAC[1][VX]*pointCT[1][VY] - unitAC[1][VY]*pointCT[1][VX]),
      (unitBC[1][VX]*pointCT[1][VY] - unitBC[1][VY]*pointCT[1][VX]),
      (unitAC[2][VX]*pointCT[2][VY] - unitAC[2][VY]*pointCT[2][VX]),
      (unitBC[2][VX]*pointCT[2][VY] - unitBC[2][VY]*pointCT[2][VX]));

   // build coefficient matrix
   Matrix6 matrix (row0, row1, row2, row3, row4, row5);

   // invert coefficient matrix
   Matrix6 matrixInverse = matrix.inverse();

   // extract solution coefficient rows
   Vector6 row0Inverse = matrixInverse[0];
   Vector6 row1Inverse = matrixInverse[1];
   Vector6 row2Inverse = matrixInverse[2];
   Vector6 row3Inverse = matrixInverse[3];
   Vector6 row4Inverse = matrixInverse[4];
   Vector6 row5Inverse = matrixInverse[5];
      
   // compute forces
   forceAC[0] = 
      row0Inverse[0]*cutterForce[VX] + 
      row0Inverse[1]*cutterForce[VY] + 
      row0Inverse[2]*cutterForce[VZ];

   forceBC[0] = 
      row1Inverse[0]*cutterForce[VX] + 
      row1Inverse[1]*cutterForce[VY] + 
      row1Inverse[2]*cutterForce[VZ];

   forceAC[1] = 
      row2Inverse[0]*cutterForce[VX] + 
      row2Inverse[1]*cutterForce[VY] + 
      row2Inverse[2]*cutterForce[VZ];

   forceBC[1] = 
      row3Inverse[0]*cutterForce[VX] + 
      row3Inverse[1]*cutterForce[VY] + 
      row3Inverse[2]*cutterForce[VZ];

   forceAC[2] = 
      row4Inverse[0]*cutterForce[VX] + 
      row4Inverse[1]*cutterForce[VY] + 
      row4Inverse[2]*cutterForce[VZ];

   forceBC[2] = 
      row5Inverse[0]*cutterForce[VX] + 
      row5Inverse[1]*cutterForce[VY] + 
      row5Inverse[2]*cutterForce[VZ];

/* 
   // validity checks - all results should be zero, and they are
   // uncomment this section to print results to standard output
   
   double result1, result2, result3, result4, result5, result6;

   result1 = 
      unitAC[0][VX]*forceAC[0] + unitBC[0][VX]*forceBC[0] + 
      unitAC[1][VX]*forceAC[1] + unitBC[1][VX]*forceBC[1] + 
      unitAC[2][VX]*forceAC[2] + unitBC[2][VX]*forceBC[2] - cutterForce[VX];

   result2 = 
      unitAC[0][VY]*forceAC[0] + unitBC[0][VY]*forceBC[0] + 
      unitAC[1][VY]*forceAC[1] + unitBC[1][VY]*forceBC[1] + 
      unitAC[2][VY]*forceAC[2] + unitBC[2][VY]*forceBC[2] - cutterForce[VY];

   result3 = 
      unitAC[0][VZ]*forceAC[0] + unitBC[0][VZ]*forceBC[0] + 
      unitAC[1][VZ]*forceAC[1] + unitBC[1][VZ]*forceBC[1] + 
      unitAC[2][VZ]*forceAC[2] + unitBC[2][VZ]*forceBC[2] - cutterForce[VZ];

   result4 =
      (unitAC[0][VY]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VY])*forceAC[0] +
      (unitBC[0][VY]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VY])*forceBC[0] +
      (unitAC[1][VY]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VY])*forceAC[1] +
      (unitBC[1][VY]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VY])*forceBC[1] +
      (unitAC[2][VY]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VY])*forceAC[2] +
      (unitBC[2][VY]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VY])*forceBC[2];

   result5 =
      (unitAC[0][VX]*pointCT[0][VZ] - unitAC[0][VZ]*pointCT[0][VX])*forceAC[0] +
      (unitBC[0][VX]*pointCT[0][VZ] - unitBC[0][VZ]*pointCT[0][VX])*forceBC[0] +
      (unitAC[1][VX]*pointCT[1][VZ] - unitAC[1][VZ]*pointCT[1][VX])*forceAC[1] +
      (unitBC[1][VX]*pointCT[1][VZ] - unitBC[1][VZ]*pointCT[1][VX])*forceBC[1] +
      (unitAC[2][VX]*pointCT[2][VZ] - unitAC[2][VZ]*pointCT[2][VX])*forceAC[2] +
      (unitBC[2][VX]*pointCT[2][VZ] - unitBC[2][VZ]*pointCT[2][VX])*forceBC[2];

   result6 =
      (unitAC[0][VX]*pointCT[0][VY] - unitAC[0][VY]*pointCT[0][VX])*forceAC[0] +
      (unitBC[0][VX]*pointCT[0][VY] - unitBC[0][VY]*pointCT[0][VX])*forceBC[0] +
      (unitAC[1][VX]*pointCT[1][VY] - unitAC[1][VY]*pointCT[1][VX])*forceAC[1] +
      (unitBC[1][VX]*pointCT[1][VY] - unitBC[1][VY]*pointCT[1][VX])*forceBC[1] +
      (unitAC[2][VX]*pointCT[2][VY] - unitAC[2][VY]*pointCT[2][VX])*forceAC[2] +
      (unitBC[2][VX]*pointCT[2][VY] - unitBC[2][VY]*pointCT[2][VX])*forceBC[2];

   cout 
      << floor(result1 + 0.5) << ' '
      << floor(result2 + 0.5) << ' '
      << floor(result3 + 0.5) << ' '
      << floor(result4 + 0.5) << ' '
      << floor(result5 + 0.5) << ' '
      << floor(result6 + 0.5) << ' ';

*/

   // return force - for now the absolute min or max
   double force = 0;  // = 0 avoids spurious warning
   switch (selector)
      {
         case MIN:
            force = fabs(forceAC[0]);
            if (force > fabs(forceBC[0]))
               force = fabs(forceBC[0]);
            if (force > fabs(forceAC[1]))
               force = fabs(forceAC[1]);
            if (force > fabs(forceBC[1]))
               force = fabs(forceBC[1]);
            if (force > fabs(forceAC[2]))
               force = fabs(forceAC[2]);
            if (force > fabs(forceBC[2]))
               force = fabs(forceBC[2]);
            break;
   
         case MAX:
            force = fabs(forceAC[0]);
            if (force < fabs(forceBC[0]))
               force = fabs(forceBC[0]);
            if (force < fabs(forceAC[1]))
               force = fabs(forceAC[1]);
            if (force < fabs(forceBC[1]))
               force = fabs(forceBC[1]);
            if (force < fabs(forceAC[2]))
               force = fabs(forceAC[2]);
            if (force < fabs(forceBC[2]))
               force = fabs(forceBC[2]);
            break;
      }
   return force;
}

double
HexapodMechanism::cutterDisplacement()
/* cutter displacement - compute leg forces first by call of legForce

      * three leg pair triangles 1, 2 and 3;
      * top apexes A and B - top vector is AB;
      * bottom apex C - legs are AC and BC;
      * point D on AB such that DC is normal to AB;
      * AD = (sqr(AB) + sqr(AC) - sqr(BC))/2AB;
      * DC = sqrt(sqr(AC) - sqr(AD));

   Strategy:
      * compute leg extensions;
      * maintaining present leg pair triangle planes,
        adjust C exactly to satisfy these leg extensions;
      * achieve this using the pythagorean equations above,
        and vector manipulations;
      * now the hard part - rotate the leg pair triangles
        to satisfy the known fixed spacings between the
        platform universal joints. 
      * in general, no analytical solution because of non-linearity.
      * however, in this case involving very small changes, a
        linear solution is available. Compute the changes made
        to the joint spacings by unit movements along the
        normals CG. With this information set up and solve
        three linear equations in three unknowns to obtain
        the actual normal movements required.
      * adjust points C normal to leg pair planes.
*/
{
   Vector3 lineAC[3], lineBC[3];  // leg vectors
   Vector3 lineAB[3];  // top vectors
   Vector3 unitDC[3];  // in-plane normal to top vector
   Vector3 unitCG[3];  // unit out-of-plane normals at C
   Vector3 lineIJ, lineJK, lineKI;  // lines joining platform pivot points 
   Vector3 unitIJ, unitJK, unitKI;

   Vector3 newPointC[3], pointD[3];  // points to be computed

   double 
      lengthAB[3], lengthAC[3], lengthBC[3],   // line lengths
      lengthAD[3], lengthDC[3],
      lengthIJ, lengthJK, lengthKI;
          
   double extensionAC[3], extensionBC[3];  // leg extensions under axial forces

   for (int n = 0; n < 3; n++)
      {
         // compute initial vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
      }

   lineIJ = pointC[1] - pointC[0];
   lineJK = pointC[2] - pointC[1];
   lineKI = pointC[0] - pointC[2];
   lengthIJ = lineIJ.length();  // need lengths, don't use full vectors
   lengthJK = lineJK.length();
   lengthKI = lineKI.length();

   unitIJ = lineIJ;
   unitJK = lineJK;
   unitKI = lineKI;
   unitIJ.normalize();  // need unit vectors
   unitJK.normalize();
   unitKI.normalize();

   for (int n = 0; n < 3; n++)
      // compute unit normal vectors at C by cross product
      unitCG[n] = (lineAC[n] ^ lineBC[n]).normalize();

   // compute their projections onto lines IJ, JK, KI
   // using dot product (p = A*B where B is a unit vector)
   double 
      projectionIJ, projectionJI, 
      projectionJK, projectionKJ,
      projectionKI, projectionIK;
   projectionIK = unitCG[0]*(-unitKI);  // perhaps signs don't matter
   projectionIJ = unitCG[0]*unitIJ;
   projectionJI = unitCG[1]*(-unitIJ);
   projectionJK = unitCG[1]*unitJK;
   projectionKJ = unitCG[2]*(-unitJK);
   projectionKI = unitCG[2]*unitKI;

   // above for use later.....

   Vector3 unitAB[3];

   for (int n = 0; n < 3; n++)
      {
         lineAB[n] = pointB[n] - pointA[n];

         lengthAB[n] = lineAB[n].length();
         lengthAC[n] = lineAC[n].length();
         lengthBC[n] = lineBC[n].length();

         lengthAD[n] = 
	   (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n]
               - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute unit vectors AB and DC
         unitAB[n] = lineAB[n];
         unitAB[n].normalize();

         pointD[n] = pointA[n] + unitAB[n]*lengthAD[n];

         unitDC[n] = (pointC[n] - pointD[n]).normalize();
      }

   // now repeat with leg lengths including the load extensions,
   // to recompute points C (still on the original planes)

   // first compute the extensions using 2 cross sectional areas, 
   // for screw and tube + end sections - could be more precise if critical
   double E = 200000;  // N/mm2 (steel)
   double 
      effectiveTubeArea, effectiveTubeLength, 
      effectiveScrewArea,
      tubeOD, tubeID, screwOD;

   tubeOD = legParameters.nutTubeOD;
   tubeID = legParameters.nutTubeID;
   effectiveTubeArea = 0.8*PI*(tubeOD*tubeOD - tubeID*tubeID)/4;  
                    // 0.8 factor for end sections

   screwOD = legParameters.screwDiameter;
   effectiveScrewArea = PI*((0.9*screwOD)*(0.9*screwOD))/4;  
                          // 0.9 factor for thread

   effectiveTubeLength = 
      legParameters.screwTopLength + 
      legParameters.nutTubeLength + 
      legParameters.nutTubeBottomLength;

   for (int n = 0; n < 3; n++)
      {
         extensionAC[n] = forceAC[n]*effectiveTubeLength/(effectiveTubeArea*E);
         extensionBC[n] = forceBC[n]*effectiveTubeLength/(effectiveTubeArea*E);

         extensionAC[n] += 
	    forceAC[n]*(lengthAC[n]
               - effectiveTubeLength)/(effectiveScrewArea*E);
         extensionBC[n] += 
	    forceBC[n]*(lengthBC[n]
               - effectiveTubeLength)/(effectiveScrewArea*E);

         double backlash = legParameters.totalBacklash/2;  
                                                // each side of average

         if (extensionAC[n] > n)
            extensionAC[n] += backlash;
         else
            extensionAC[n] -= backlash;
            
         if (extensionBC[n] > n)
            extensionBC[n] += backlash;
         else
            extensionBC[n] -= backlash;

         // alter leg lengths
         lengthAC[n] += extensionAC[n];
         lengthBC[n] += extensionBC[n];

         // recompute lengths AD
         lengthAD[n] = 
	   (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n]
               - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute lengths DC - didn't need to do this the first time
         lengthDC[n] = sqrt(lengthAC[n]*lengthAC[n] - lengthAD[n]*lengthAD[n]);

         // and now we have the new points C by vector arithmetic
         newPointC[n] = 
            pointA[n] + unitAB[n]*lengthAD[n] + unitDC[n]*lengthDC[n];
      }

   double newLengthIJ, newLengthJK, newLengthKI;  // new lengths
   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   double errorIJ, errorJK, errorKI;  // fit errors to be corrected
   errorIJ = newLengthIJ - lengthIJ;
   errorJK = newLengthJK - lengthJK;
   errorKI = newLengthKI - lengthKI;

   /* Now set up matrix to solve equations based on projections above.
      Equations are:
         errorIJ = projectionIJ*i + projectionJI*j + 0*k; 
         errorJK = 0*i + projectionJK*j + projectionKJ*k;
         errorKI = projectionIK*i + 0*j + projectionKI*k;
   */

   // build coefficient matrix rows
   Vector3 row0 (projectionIJ, projectionJI, 0);
   Vector3 row1 (0, projectionJK, projectionKJ);
   Vector3 row2 (projectionIK, 0, projectionKI);

   // build coefficient matrix
   Matrix3 matrix (row0, row1, row2);

   // invert coefficient matrix
   Matrix3 matrixInverse = matrix.inverse();

   // extract solution coefficient rows
   Vector3 row0Inverse = matrixInverse[0];
   Vector3 row1Inverse = matrixInverse[1];
   Vector3 row2Inverse = matrixInverse[2];

   // compute i, j, k
   double i, j, k;
   i = row0Inverse[0]*errorIJ + row0Inverse[1]*errorJK + row0Inverse[2]*errorKI;
   j = row1Inverse[0]*errorIJ + row1Inverse[1]*errorJK + row1Inverse[2]*errorKI;
   k = row2Inverse[0]*errorIJ + row2Inverse[1]*errorJK + row2Inverse[2]*errorKI;

   // finally we arrive at adjusted points C by vector arithmetic
   newPointC[0] += unitCG[0]*i;
   newPointC[1] += unitCG[1]*j;
   newPointC[2] += unitCG[2]*k;

/*
   // check solution
   double newLengthAC[3], newLengthBC[3];

   for (int n = 0; n < 3; n++)
      {
         newLengthAC[n] = (newPointC[n] - pointA[n]).length();
         newLengthBC[n] = (newPointC[n] - pointB[n]).length();
      }

   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   // results should all be zero, and they are
   cout << floor(1000*(newLengthAC[0] - lengthAC[0]) + 0.5) << ' '
        << floor(1000*(newLengthAC[1] - lengthAC[1]) + 0.5) << ' '
        << floor(1000*(newLengthAC[2] - lengthAC[2]) + 0.5) << ' '
        << floor(1000*(newLengthBC[0] - lengthBC[0]) + 0.5) << ' '
        << floor(1000*(newLengthBC[1] - lengthBC[1]) + 0.5) << ' '
        << floor(1000*(newLengthBC[2] - lengthBC[2]) + 0.5) << ' '
        << floor(1000*(newLengthIJ - lengthIJ) + 0.5) << ' '
        << floor(1000*(newLengthJK - lengthJK) + 0.5) << ' ' 
        << floor(1000*(newLengthKI - lengthKI) + 0.5) << ' ';
*/

   Vector3 pointN;  // midpoint of JK
   Vector3 pointP, newPointP;  // cutter location
   Vector3 lineIM;  // bisector of angleJIK, length 2/3 of IN, locating centroid
   Vector3 lineIK, newLineIJ, newLineIK;  // after displacement
   Vector3 unitIL;  // normal to platform plane
   Vector3 lineMP;  // normal to platform from M, length for now 150

   // first, where we should be
   pointN = (pointC[1] + pointC[2])/2;
   lineIM = (pointN - pointC[0])*2/3;
   lineIK = -lineKI;
   unitIL = (lineIJ ^ lineIK).normalize();
   lineMP = -unitIL*150;  // sign is correct
   pointP = pointC[0] + lineIM + lineMP;

   // then, where we actually are
   pointN = (newPointC[1] + newPointC[2])/2;
   lineIM = (pointN - newPointC[0])*2/3;
   newLineIJ = newPointC[1] - newPointC[0];
   newLineIK = newPointC[2] - newPointC[0];
   unitIL = (newLineIJ ^ newLineIK).normalize();
   lineMP = -unitIL*150;  // sign is correct
   newPointP = newPointC[0] + lineIM + lineMP;

   /* All the above checked with ddd and by examining output for 
      different configurations. In particular the signs are correct, 
      and the displacement is properly propagated to the cutter at 
      point P which is in the right place. The validity check
      shows that the solution is correct. */

   double displacement;
   displacement = floor(1000*(newPointP - pointP).length());  // microns
   // displacement = floor(1000*(newPointP - pointP)[VX]);
   // displacement = floor(1000*(newPointP - pointP)[VY]);
   // displacement = floor(1000*(newPointP - pointP)[VZ]);
   return displacement;
}

double
HexapodMechanism::resolutionError()
// compute resolution error, based closely on cutterDisplacement()
{
   Vector3 lineAC[3], lineBC[3];  // leg vectors
   Vector3 lineAB[3];  // top vectors
   Vector3 unitDC[3];  // in-plane normal to top vector
   Vector3 unitCG[3];  // unit out-of-plane normals at C
   Vector3 lineIJ, lineJK, lineKI;  // lines joining platform pivot points 
   Vector3 unitIJ, unitJK, unitKI;

   Vector3 newPointC[3], pointD[3];  // points to be computed

   double 
      lengthAB[3], lengthAC[3], lengthBC[3],  // line lengths
      lengthAD[3], lengthDC[3],
      lengthIJ, lengthJK, lengthKI;
	  
   double resolution = 0.0125;  // leg step resolution;
   double correctionAC[3], correctionBC[3];  // leg length corrections 
   double initialLength, correctedLength;  // leg lengths

   // compute leg length corrections resulting from resolution granularity 
   for (int i = 0; i < 2; i++)  // loop over the two legs of a pair
      for (int j = 0; j < 3; j++)  // loop over the three leg pairs
	 switch (i)
	    {
	       case 0:  // leg AC[j]
                  initialLength = (pointC[j] - pointA[j]).length();
                  correctedLength = floor(initialLength/resolution)*resolution;
	          correctionAC[j] = correctedLength - initialLength;
	          break;
	       case 1:  // leg BC[j]
                  initialLength = (pointC[j] - pointB[j]).length();
                  correctedLength = floor(initialLength/resolution)*resolution;
	          correctionBC[j] = correctedLength - initialLength;
	          break;
	    }

   // now compute the changes in platform joint position
   // resulting from these leg length corrections
	    
   // ********************************************************************
   //          code taken from cutterDisplacement() follows
   // ********************************************************************

   for (int n = 0; n < 3; n++)
      {
         // compute initial vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
      }

   lineIJ = pointC[1] - pointC[0];
   lineJK = pointC[2] - pointC[1];
   lineKI = pointC[0] - pointC[2];
   lengthIJ = lineIJ.length();  // need lengths, don't use full vectors
   lengthJK = lineJK.length();
   lengthKI = lineKI.length();

   unitIJ = lineIJ;
   unitJK = lineJK;
   unitKI = lineKI;
   unitIJ.normalize();  // need unit vectors
   unitJK.normalize();
   unitKI.normalize();

   for (int n = 0; n < 3; n++)
      // compute unit normal vectors at C by cross product
      unitCG[n] = (lineAC[n] ^ lineBC[n]).normalize();

   // compute their projections onto lines IJ, JK, KI
   // using dot product (p = A*B where B is a unit vector)
   double 
      projectionIJ, projectionJI, 
      projectionJK, projectionKJ,
      projectionKI, projectionIK;
   projectionIK = unitCG[0]*(-unitKI);  // perhaps signs don't matter
   projectionIJ = unitCG[0]*unitIJ;
   projectionJI = unitCG[1]*(-unitIJ);
   projectionJK = unitCG[1]*unitJK;
   projectionKJ = unitCG[2]*(-unitJK);
   projectionKI = unitCG[2]*unitKI;

   // above for use later.....

   Vector3 unitAB[3];

   for (int n = 0; n < 3; n++)
      {
         lineAB[n] = pointB[n] - pointA[n];

         lengthAB[n] = lineAB[n].length();
         lengthAC[n] = lineAC[n].length();
         lengthBC[n] = lineBC[n].length();

         lengthAD[n] = 
	    (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n] 
	        - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute unit vectors AB and DC
         unitAB[n] = lineAB[n];
         unitAB[n].normalize();

         pointD[n] = pointA[n] + unitAB[n]*lengthAD[n];

         unitDC[n] = (pointC[n] - pointD[n]).normalize();
      }

   // now repeat with corrected leg lengths to 
   // recompute points C (still on the original planes)
   
   for (int n = 0; n < 3; n++)
      {
         // ***************************************
	 //  changes to cutterDisplacement() code
         // ***************************************

         // alter leg lengths
         lengthAC[n] -= correctionAC[n];
         lengthBC[n] -= correctionBC[n];

         // ***************************************
	 //             end changes 
         // ***************************************

         // recompute lengths AD
         lengthAD[n] = 
	    (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n] 
	        - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute lengths DC - didn't need to do this the first time
         lengthDC[n] = sqrt(lengthAC[n]*lengthAC[n] - lengthAD[n]*lengthAD[n]);

         // and now we have the new points C by vector arithmetic
         newPointC[n] = 
	    pointA[n] + unitAB[n]*lengthAD[n] + unitDC[n]*lengthDC[n];
      }

   double newLengthIJ, newLengthJK, newLengthKI;  // new lengths
   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   double errorIJ, errorJK, errorKI;  // fit errors to be corrected
   errorIJ = newLengthIJ - lengthIJ;
   errorJK = newLengthJK - lengthJK;
   errorKI = newLengthKI - lengthKI;

   /* Now set up matrix to solve equations based on projections above.
      Equations are:
         errorIJ = projectionIJ*i + projectionJI*j + 0*k; 
         errorJK = 0*i + projectionJK*j + projectionKJ*k;
         errorKI = projectionIK*i + 0*j + projectionKI*k;
   */

   // build coefficient matrix rows
   Vector3 row0 (projectionIJ, projectionJI, 0);
   Vector3 row1 (0, projectionJK, projectionKJ);
   Vector3 row2 (projectionIK, 0, projectionKI);

   // build coefficient matrix
   Matrix3 matrix (row0, row1, row2);

   // invert coefficient matrix
   Matrix3 matrixInverse = matrix.inverse();

   // extract solution coefficient rows
   Vector3 row0Inverse = matrixInverse[0];
   Vector3 row1Inverse = matrixInverse[1];
   Vector3 row2Inverse = matrixInverse[2];

   { 
      // g++ bug requires this to be in a block of its own -
      // the variable i gets exported as global
      // I could rename the variables, but that would
      // upset the nomenclature
      double i, j, k;
      i = row0Inverse[0]*errorIJ 
             + row0Inverse[1]*errorJK + row0Inverse[2]*errorKI;
      j = row1Inverse[0]*errorIJ 
             + row1Inverse[1]*errorJK + row1Inverse[2]*errorKI;
      k = row2Inverse[0]*errorIJ 
             + row2Inverse[1]*errorJK + row2Inverse[2]*errorKI;

      // finally we arrive at adjusted points C by vector arithmetic
      newPointC[0] += unitCG[0]*i;
      newPointC[1] += unitCG[1]*j;
      newPointC[2] += unitCG[2]*k;
   }

/*
   // check solution
   double newLengthAC[3], newLengthBC[3];

   for (int n = 0; n < 3; n++)
      {
         newLengthAC[n] = (newPointC[n] - pointA[n]).length();
         newLengthBC[n] = (newPointC[n] - pointB[n]).length();
      }

   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   // results should all be zero, and they are
   cout 
      << floor(1000*(newLengthAC[0] - lengthAC[0]) + 0.5) << ' '
      << floor(1000*(newLengthAC[1] - lengthAC[1]) + 0.5) << ' '
      << floor(1000*(newLengthAC[2] - lengthAC[2]) + 0.5) << ' '
      << floor(1000*(newLengthBC[0] - lengthBC[0]) + 0.5) << ' '
      << floor(1000*(newLengthBC[1] - lengthBC[1]) + 0.5) << ' '
      << floor(1000*(newLengthBC[2] - lengthBC[2]) + 0.5) << ' '
      << floor(1000*(newLengthIJ - lengthIJ) + 0.5) << ' '
      << floor(1000*(newLengthJK - lengthJK) + 0.5) << ' ' 
      << floor(1000*(newLengthKI - lengthKI) + 0.5) << ' ';
*/

   Vector3 pointN;  // midpoint of JK
   Vector3 pointP, newPointP;  // cutter location
   Vector3 lineIM;  // bisector of angleJIK, length 2/3 of IN, locating centroid
   Vector3 lineIK, newLineIJ, newLineIK;  // after displacement
   Vector3 unitIL;  // normal to platform plane
   Vector3 lineMP;  // normal to platform from M, length for now 150

   // first, where we should be
   pointN = (pointC[1] + pointC[2])/2;
   lineIM = (pointN - pointC[0])*2/3;
   lineIK = -lineKI;
   unitIL = (lineIJ ^ lineIK).normalize();
   lineMP = -unitIL*150;  // sign is correct
   pointP = pointC[0] + lineIM + lineMP;

   // then, where we actually are
   pointN = (newPointC[1] + newPointC[2])/2;
   lineIM = (pointN - newPointC[0])*2/3;
   newLineIJ = newPointC[1] - newPointC[0];
   newLineIK = newPointC[2] - newPointC[0];
   unitIL = (newLineIJ ^ newLineIK).normalize();
   lineMP = -unitIL*150;  // sign is correct
   newPointP = newPointC[0] + lineIM + lineMP;

   /* All the above checked with ddd and by examining output for 
      different configurations. In particular the signs are correct, 
      and the displacement is properly propagated to the cutter at 
      point P which is in the right place. The validity check
      shows that the solution is correct. */

   // ********************************************************************
   //          end of code taken from cutterDisplacement()
   // ********************************************************************

   // return error as a multiple of the resolution
   double error;
   error = ceil((newPointP - pointP).length()/resolution);
   return error;
}

void
HexapodMechanism::computeCalibrationCoefficients()
/* compute calibration coefficients, based closely on cutterDisplacement()

   We want to establish the effect at calibration points of small changes in 
   base geometry. Though the mechanism is non-linear, in a given position the
   effect of small changes will be linear. Thus we can look at the possible
   corrections (eg dx1, dy1, dz1) one at a time, computing the linear 
   coefficient relating a single change (eg dx1) to a calibration measurement
   (eg dz at a platform joint). When we have the set of coefficients, we can
   write the equation (more precisely the inequalities) defining constraints
   on the correction variables applying at that calibration point. So in this
   function, we look in turn at each possible correction, and find the
   coefficient. This problem is similar to the cutter displacement problem, 
   where we have to find the change in geometry resulting from a change in
   leg length under load. In this first approach to the problem, we consider 
   only the effect of changes in top joint coordinates. We move a top joint
   a small amount in an axial direction, and calculate the change in leg length
   which that would produce if the platform remained in the same position. We
   then use the method developed for cutter displacement calculations to
   compute the change in the platform position which this change in leg length
   would produce. This gives us the coefficients we need.

   Nomenclature:
      * three leg pair triangles 1, 2 and 3;
      * top apexes A and B - top vector is AB;
      * bottom apex C - legs are AC and BC;
      * point D on AB such that DC is normal to AB;
      * AD = (sqr(AB) + sqr(AC) - sqr(BC))/2AB;
      * DC = sqrt(sqr(AC) - sqr(AD));

   Strategy:
      * consider each possible correction in turn;
      * compute change in leg length if platform remains in original position;
      * maintaining present leg pair triangle planes,
        adjust C exactly to correct this leg length change which cannot occur;
      * achieve this using the pythagorean equations above,
        and vector manipulations;
      * now the hard part - rotate the leg pair triangles
        to satisfy the known fixed spacings between the
        platform universal joints. 
      * in general, no analytical solution because of non-linearity.
      * however, in this case involving very small changes, a
        linear solution is available. Compute the changes made
        to the joint spacings by unit movements along the
        normals CG. With this information set up and solve
        three linear equations in three unknowns to obtain
        the actual normal movements required.
      * adjust points C normal to leg pair planes;
      * compute required correction coefficients.
*/
{
   Vector3 lineAC[3], lineBC[3];  // leg vectors
   Vector3 lineAB[3];  // top vectors
   Vector3 unitDC[3];  // in-plane normal to top vector
   Vector3 unitCG[3];  // unit out-of-plane normals at C
   Vector3 lineIJ, lineJK, lineKI;  // lines joining platform pivot points 
   Vector3 unitIJ, unitJK, unitKI;

   Vector3 pointA[3], pointB[3];  // *** local variables hiding globals ***
   Vector3 newPointC[3], pointD[3];  // points to be computed

   double 
      lengthAB[3], lengthAC[3], lengthBC[3],   // line lengths
      lengthAD[3], lengthDC[3],
      lengthIJ, lengthJK, lengthKI;
          
   double delta;  // small correction to individual mechanism variables
   double correctionAC[3], correctionBC[3];  // leg length corrections 
   // platform joint spacing corrections
   double correctionIJ, correctionJK, correctionKI;  
   double initialLength, correctedLength;  // leg lengths
   delta = 0.01;  // nominal small amount

   // loop through all considered corrections, calculating
   // the calibration coefficient for each separately
   for (int i = 0; i < correctionCount; i++)
      {
         for (int n = 0; n < 3; n++)
            {
               // copy global data to local
               pointA[n] = this->pointA[n];
               pointB[n] = this->pointB[n];

               // zero leg length corrections
               correctionAC[n] = 0.0;
               correctionBC[n] = 0.0;
            }

         // zero platform joint spacing corrections
	 correctionIJ = 0.0;
	 correctionJK = 0.0;
	 correctionKI = 0.0;

         switch (i)
            {
               case 0:  // pointA[0].x
                  initialLength = (pointC[0] - pointA[0]).length();
                  pointA[0][VX] += delta;
                  correctedLength = (pointC[0] - pointA[0]).length();
                  correctionAC[0] = correctedLength - initialLength;
                  break;
               case 1:  // pointA[0].y
                  initialLength = (pointC[0] - pointA[0]).length();
                  pointA[0][VY] += delta;
                  correctedLength = (pointC[0] - pointA[0]).length();
                  correctionAC[0] = correctedLength - initialLength;
                  break;
               case 2:  // pointA[0].z
                  initialLength = (pointC[0] - pointA[0]).length();
                  pointA[0][VZ] += delta;
                  correctedLength = (pointC[0] - pointA[0]).length();
                  correctionAC[0] = correctedLength - initialLength;
                  break;

               case 3:  // pointA[1].x
                  initialLength = (pointC[1] - pointA[1]).length();
                  pointA[1][VX] += delta;
                  correctedLength = (pointC[1] - pointA[1]).length();
                  correctionAC[1] = correctedLength - initialLength;
                  break;
               case 4:  // pointA[1].y
                  initialLength = (pointC[1] - pointA[1]).length();
                  pointA[1][VY] += delta;
                  correctedLength = (pointC[1] - pointA[1]).length();
                  correctionAC[1] = correctedLength - initialLength;
                  break;
               case 5:  // pointA[1].z
                  initialLength = (pointC[1] - pointA[1]).length();
                  pointA[1][VZ] += delta;
                  correctedLength = (pointC[1] - pointA[1]).length();
                  correctionAC[1] = correctedLength - initialLength;
                  break;

               case 6:  // pointA[2].x
                  initialLength = (pointC[2] - pointA[2]).length();
                  pointA[2][VX] += delta;
                  correctedLength = (pointC[2] - pointA[2]).length();
                  correctionAC[2] = correctedLength - initialLength;
                  break;
               case 7:  // pointA[2].y
                  initialLength = (pointC[2] - pointA[2]).length();
                  pointA[2][VY] += delta;
                  correctedLength = (pointC[2] - pointA[2]).length();
                  correctionAC[2] = correctedLength - initialLength;
                  break;
               case 8:  // pointA[2].z
                  initialLength = (pointC[2] - pointA[2]).length();
                  pointA[2][VZ] += delta;
                  correctedLength = (pointC[2] - pointA[2]).length();
                  correctionAC[2] = correctedLength - initialLength;
                  break;

               case 9:  // pointB[0].x
                  initialLength = (pointC[0] - pointB[0]).length();
                  pointB[0][VX] += delta;
                  correctedLength = (pointC[0] - pointB[0]).length();
                  correctionBC[0] = correctedLength - initialLength;
                  break;
               case 10:  // pointB[0].y
                  initialLength = (pointC[0] - pointB[0]).length();
                  pointB[0][VY] += delta;
                  correctedLength = (pointC[0] - pointB[0]).length();
                  correctionBC[0] = correctedLength - initialLength;
                  break;
               case 11:  // pointB[0].z
                  initialLength = (pointC[0] - pointB[0]).length();
                  pointB[0][VZ] += delta;
                  correctedLength = (pointC[0] - pointB[0]).length();
                  correctionBC[0] = correctedLength - initialLength;
                  break;

               case 12:  // pointB[1].x
                  initialLength = (pointC[1] - pointB[1]).length();
                  pointB[1][VX] += delta;
                  correctedLength = (pointC[1] - pointB[1]).length();
                  correctionBC[1] = correctedLength - initialLength;
                  break;
               case 13:  // pointB[1].y
                  initialLength = (pointC[1] - pointB[1]).length();
                  pointB[1][VY] += delta;
                  correctedLength = (pointC[1] - pointB[1]).length();
                  correctionBC[1] = correctedLength - initialLength;
                  break;
               case 14:  // pointB[1].z
                  initialLength = (pointC[1] - pointB[1]).length();
                  pointB[1][VZ] += delta;
                  correctedLength = (pointC[1] - pointB[1]).length();
                  correctionBC[1] = correctedLength - initialLength;
                  break;

               case 15:  // pointB[2].x
                  initialLength = (pointC[2] - pointB[2]).length();
                  pointB[2][VX] += delta;
                  correctedLength = (pointC[2] - pointB[2]).length();
                  correctionBC[2] = correctedLength - initialLength;
                  break;
               case 16:  // pointB[2].y
                  initialLength = (pointC[2] - pointB[2]).length();
                  pointB[2][VY] += delta;
                  correctedLength = (pointC[2] - pointB[2]).length();
                  correctionBC[2] = correctedLength - initialLength;
                  break;
               case 17:  // pointB[2].z
                  initialLength = (pointC[2] - pointB[2]).length();
                  pointB[2][VZ] += delta;
                  correctedLength = (pointC[2] - pointB[2]).length();
                  correctionBC[2] = correctedLength - initialLength;
                  break;

	       case 18:  // lengthA[0]
	          correctionAC[0] = -delta;
	          break;
	       case 19:  // lengthA[1]
	          correctionAC[1] = -delta;
	          break;
	       case 20:  // lengthA[2]
	          correctionAC[2] = -delta;
	          break;
	       case 21:  // lengthB[0]
	          correctionBC[0] = -delta;
	          break;
	       case 22:  // lengthB[1]
	          correctionBC[1] = -delta;
	          break;
	       case 23:  // lengthB[2]
	          correctionBC[2] = -delta;
	          break;

	       case 24:  // lengthIJ (C0-C1)
	          correctionIJ = delta;
	          break;
	       case 25:  // lengthJK (C1-C2)
	          correctionJK = delta;
	          break;
	       case 26:  // lengthKI (C2-C0)
	          correctionKI = delta;
	          break;
            }

   // ********************************************************************
   //          code taken from cutterDisplacement() follows
   // ********************************************************************

   for (int n = 0; n < 3; n++)
      {
         // compute initial vectors
         lineAC[n] = pointC[n] - pointA[n];
         lineBC[n] = pointC[n] - pointB[n];
      }

   lineIJ = pointC[1] - pointC[0];
   lineJK = pointC[2] - pointC[1];
   lineKI = pointC[0] - pointC[2];
   lengthIJ = lineIJ.length();  // need lengths, don't use full vectors
   lengthJK = lineJK.length();
   lengthKI = lineKI.length();

   unitIJ = lineIJ;
   unitJK = lineJK;
   unitKI = lineKI;
   unitIJ.normalize();  // need unit vectors
   unitJK.normalize();
   unitKI.normalize();

   for (int n = 0; n < 3; n++)
      // compute unit normal vectors at C by cross product
      unitCG[n] = (lineAC[n] ^ lineBC[n]).normalize();

   // compute their projections onto lines IJ, JK, KI
   // using dot product (p = A*B where B is a unit vector)
   double 
      projectionIJ, projectionJI, 
      projectionJK, projectionKJ,
      projectionKI, projectionIK;
   projectionIK = unitCG[0]*(-unitKI);  // perhaps signs don't matter
   projectionIJ = unitCG[0]*unitIJ;
   projectionJI = unitCG[1]*(-unitIJ);
   projectionJK = unitCG[1]*unitJK;
   projectionKJ = unitCG[2]*(-unitJK);
   projectionKI = unitCG[2]*unitKI;

   // above for use later.....

   Vector3 unitAB[3];

   for (int n = 0; n < 3; n++)
      {
         lineAB[n] = pointB[n] - pointA[n];

         lengthAB[n] = lineAB[n].length();
         lengthAC[n] = lineAC[n].length();
         lengthBC[n] = lineBC[n].length();

         lengthAD[n] = 
	    (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n] 
	        - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute unit vectors AB and DC
         unitAB[n] = lineAB[n];
         unitAB[n].normalize();

         pointD[n] = pointA[n] + unitAB[n]*lengthAD[n];

         unitDC[n] = (pointC[n] - pointD[n]).normalize();
      }

   // now repeat with corrected leg lengths to 
   // recompute points C (still on the original planes)
   
   for (int n = 0; n < 3; n++)
      {
         // ***************************************
         //  changes to cutterDisplacement() code
         // ***************************************

         // alter leg lengths - only one will change
         lengthAC[n] -= correctionAC[n];
         lengthBC[n] -= correctionBC[n];

         // ***************************************
         //             end changes 
         // ***************************************

         // recompute lengths AD
         lengthAD[n] = 
	    (lengthAB[n]*lengthAB[n] + lengthAC[n]*lengthAC[n] 
	        - lengthBC[n]*lengthBC[n])/(2*lengthAB[n]);

         // compute lengths DC - didn't need to do this the first time
         lengthDC[n] = sqrt(lengthAC[n]*lengthAC[n] - lengthAD[n]*lengthAD[n]);

         // and now we have the new points C by vector arithmetic
         newPointC[n] = 
            pointA[n] + unitAB[n]*lengthAD[n] + unitDC[n]*lengthDC[n];
      }

   double newLengthIJ, newLengthJK, newLengthKI;  // new lengths
   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   // *************************************************************
   //             changes to cutterDisplacement() code
   // *************************************************************

   double errorIJ, errorJK, errorKI;  // fit errors to be corrected
   errorIJ = newLengthIJ - lengthIJ - correctionIJ;
   errorJK = newLengthJK - lengthJK - correctionJK;
   errorKI = newLengthKI - lengthKI - correctionKI;

   // *************************************************************
   //                        end changes 
   // *************************************************************


   /* Now set up matrix to solve equations based on projections above.
      Equations are:
         errorIJ = projectionIJ*i + projectionJI*j + 0*k; 
         errorJK = 0*i + projectionJK*j + projectionKJ*k;
         errorKI = projectionIK*i + 0*j + projectionKI*k;
   */

   // build coefficient matrix rows
   Vector3 row0 (projectionIJ, projectionJI, 0);
   Vector3 row1 (0, projectionJK, projectionKJ);
   Vector3 row2 (projectionIK, 0, projectionKI);

   // build coefficient matrix
   Matrix3 matrix (row0, row1, row2);

   // invert coefficient matrix
   Matrix3 matrixInverse = matrix.inverse();

   // extract solution coefficient rows
   Vector3 row0Inverse = matrixInverse[0];
   Vector3 row1Inverse = matrixInverse[1];
   Vector3 row2Inverse = matrixInverse[2];

   // compute i, j, k, in a separate block because those names 
   // have been used in the enclosing context for loop variables!
   {
      double i, j, k;
      i = row0Inverse[0]*errorIJ 
             + row0Inverse[1]*errorJK + row0Inverse[2]*errorKI;
      j = row1Inverse[0]*errorIJ 
             + row1Inverse[1]*errorJK + row1Inverse[2]*errorKI;
      k = row2Inverse[0]*errorIJ 
             + row2Inverse[1]*errorJK + row2Inverse[2]*errorKI;

      // finally we arrive at adjusted points C by vector arithmetic
      newPointC[0] += unitCG[0]*i;
      newPointC[1] += unitCG[1]*j;
      newPointC[2] += unitCG[2]*k;
   }
   
   /*
   // check solution - remove enclosing comment brackets if required

   double newLengthAC[3], newLengthBC[3];

   for (int n = 0; n < 3; n++)
      {
         newLengthAC[n] = (newPointC[n] - pointA[n]).length();
         newLengthBC[n] = (newPointC[n] - pointB[n]).length();
      }

   newLengthIJ = (newPointC[1] - newPointC[0]).length();
   newLengthJK = (newPointC[2] - newPointC[1]).length();
   newLengthKI = (newPointC[0] - newPointC[2]).length();

   // results should all be zero, and they are
   cout 
      << floor(1000*(newLengthAC[0] - lengthAC[0]) + 0.5) << ' '
      << floor(1000*(newLengthAC[1] - lengthAC[1]) + 0.5) << ' '
      << floor(1000*(newLengthAC[2] - lengthAC[2]) + 0.5) << ' '
      << floor(1000*(newLengthBC[0] - lengthBC[0]) + 0.5) << ' '
      << floor(1000*(newLengthBC[1] - lengthBC[1]) + 0.5) << ' '
      << floor(1000*(newLengthBC[2] - lengthBC[2]) + 0.5) << ' '
      << floor(1000*(newLengthIJ - lengthIJ) + 0.5) << ' '
      << floor(1000*(newLengthJK - lengthJK) + 0.5) << ' ' 
      << floor(1000*(newLengthKI - lengthKI) + 0.5) << ' ';

   // end of check solution
   */

   // ********************************************************************
   //          end of code taken from cutterDisplacement()
   // ********************************************************************

         // record coefficients
         for (int n = 0; n < 3; n++)
            coefficients[i][n] = (newPointC[n] - pointC[n])/delta;

      }  // end of loop over all the considered corrections

   // Further checks of the calibration coefficient set:
   // (1) x movement resulting from a general x translation will be equal to
   // the translation, so the sum of the x coefficients for a platform
   // joint should be 1.
   // (2) y movement resulting from a general x translation will be zero, so 
   // the sum of the y coefficients for a platform joint should be zero;
   // (3) the same with z;
   // Repeat for the other possible global translations.

   // this should output lines containing 0's and 1's.
   /*
   for (int n = 0; n < 3; n++)  // loop over the three platform joints
      {
         double sum;

         // without the rounding, we see significant small errors

	 // x corrections
         sum = coefficients[0][n][VX] + coefficients[3][n][VX] 
	     + coefficients[6][n][VX] + coefficients[9][n][VX] 
	     + coefficients[12][n][VX] + coefficients[15][n][VX];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[0][n][VY] + coefficients[3][n][VY] 
	     + coefficients[6][n][VY] + coefficients[9][n][VY] 
	     + coefficients[12][n][VY] + coefficients[15][n][VY];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[0][n][VZ] + coefficients[3][n][VZ] 
	     + coefficients[6][n][VZ] + coefficients[9][n][VZ] 
	     + coefficients[12][n][VZ] + coefficients[15][n][VZ];
         cout << floor(1000*sum + 0.5)/1000 << "  ";

	 // y corrections
         sum = coefficients[1][n][VX] + coefficients[4][n][VX] 
	     + coefficients[7][n][VX] + coefficients[10][n][VX] 
	     + coefficients[13][n][VX] + coefficients[16][n][VX];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[1][n][VY] + coefficients[4][n][VY] 
	     + coefficients[7][n][VY] + coefficients[10][n][VY] 
	     + coefficients[13][n][VY] + coefficients[16][n][VY];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[1][n][VZ] + coefficients[4][n][VZ] 
	     + coefficients[7][n][VZ] + coefficients[10][n][VZ] 
	     + coefficients[13][n][VZ] + coefficients[16][n][VZ];
         cout << floor(1000*sum + 0.5)/1000 << "  ";

	 // z corrections
         sum = coefficients[2][n][VX] + coefficients[5][n][VX] 
	     + coefficients[8][n][VX] + coefficients[11][n][VX] 
	     + coefficients[14][n][VX] + coefficients[17][n][VX];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[2][n][VY] + coefficients[5][n][VY] 
	     + coefficients[8][n][VY] + coefficients[11][n][VY] 
	     + coefficients[14][n][VY] + coefficients[17][n][VY];
         cout << floor(1000*sum + 0.5)/1000 << ' ';
         sum = coefficients[2][n][VZ] + coefficients[5][n][VZ] 
	     + coefficients[8][n][VZ] + coefficients[11][n][VZ] 
	     + coefficients[14][n][VZ] + coefficients[17][n][VZ];
         cout << floor(1000*sum + 0.5)/1000 << "  ";
      }
      cout << '\n';
      */
}

void
HexapodMechanism::reportRangeStatus()
{
   if (leg1A.inRange() && leg1B.inRange() &&
       leg2A.inRange() && leg2B.inRange() &&
       leg3A.inRange() && leg3B.inRange())
      rangeChecks << ". ";
   else
      {
         int count = 0;

         if (leg1A.tooShort() || leg1B.tooShort() ||
             leg2A.tooShort() || leg2B.tooShort() ||
             leg3A.tooShort() || leg3B.tooShort())
            {
               rangeChecks << 'S';
               count += 1;
            }
         if (leg1A.tooLong() || leg1B.tooLong() ||
             leg2A.tooLong() || leg2B.tooLong() ||
             leg3A.tooLong() || leg3B.tooLong())
            {
               rangeChecks << 'L';
               count += 1;
            }
         if (count == 1)
            rangeChecks << " ";
      }
}

void
HexapodMechanism::reportAngles()
{
   minAngleACB << setw(3) << floor(angleACB(MIN)*180/PI) << " ";
   maxAngleACB << setw(3) << floor(angleACB(MAX)*180/PI) << " ";

   minAngleBAC << setw(3) << floor(angleBAC(MIN)*180/PI) << " ";
   maxAngleBAC << setw(3) << floor(angleBAC(MAX)*180/PI) << " ";

   minAngleGCV << setw(3) << floor(angleGCV(MIN)*180/PI) << " ";
   maxAngleGCV << setw(3) << floor(angleGCV(MAX)*180/PI) << " ";

   minAngleACF << setw(3) << floor(angleACF(MIN)*180/PI) << " ";
   maxAngleACF << setw(3) << floor(angleACF(MAX)*180/PI) << " ";

   minAngleGCH << setw(3) << floor(angleGCH(MIN)*180/PI) << " ";
   maxAngleGCH << setw(3) << floor(angleGCH(MAX)*180/PI) << " ";
}

void
HexapodMechanism::reportForces()
{
   minLegForce << setw(3) << floor(legForce(MIN)) << " ";
   maxLegForce << setw(3) << floor(legForce(MAX)) << " ";
}

void
HexapodMechanism::reportDisplacement()
{
   displacement << setw(3) << floor(cutterDisplacement()) << " ";
}

void
HexapodMechanism::reportResolutionError()
{
   resolutionErrors << setw(3) << resolutionError() << " ";
}

void
HexapodMechanism::reportCorrections()
{
   // in the order printed by lp_solve
   for (int i = 24; i < correctionCount; i++)
      if (corrections[i] != 0.0)
         {
	    correction << "   ";
	    for (int j = 0; j < 3; j++)
	       correction << correctionName[i][j];
            correction << " = " << corrections[i] << '\n';
         }
   for (int i = 18; i < 24; i++)
      if (corrections[i] != 0.0)
         {
	    correction << "   ";
	    for (int j = 0; j < 3; j++)
	       correction << correctionName[i][j];
            correction << " = " << corrections[i] << '\n';
         }
   for (int i = 0; i < 18; i++)
      if (corrections[i] != 0.0)
         {
	    correction << "   ";
	    for (int j = 0; j < 3; j++)
	       correction << correctionName[i][j];
            correction << " = " << corrections[i] << '\n';
         }
}

void
HexapodMechanism::reportCoefficient(const int correction,
                                    const int location,
				    const int axis)
{
   // call computeCalibrationCoefficients() for the current platform position
   // before this function is called

   // report coefficients as a percentage of the correction
   double value = coefficients[correction][location][axis];
   if (value < 0.0)
      {  
         if (value > -0.01)  // don't want -0 printouts!
	    value = 0.0;
         coefficient << setw(3) << ceil(value*100) << " ";
      }
   else
      coefficient << setw(3) << floor(value*100) << " ";
}

void
HexapodMechanism::defineCorrections(const double values[correctionCount])
{
   for (int i = 0; i < correctionCount; i++)
      corrections[i] = values[i];
}

void
HexapodMechanism::generateConstraints(const int location,
				      const int axis,
				      const double maxNoise)
{
   // Generates calibration simulation constraints for lp_solve

   // call computeCalibrationCoefficients() for the current platform position
   // before this function is called

   // compute simulation constant
   const double tiny = 1.0e-10;
   double simulationConstant = 0.0;
   for (int i = 0; i < correctionCount; i++)
      if (corrections[i] != 0.0)
         simulationConstant += coefficients[i][location][axis]*corrections[i];
   // -1.0 < random < 1.0
   double random = 2.0*(rand() - RAND_MAX/2)/(double)RAND_MAX;
   simulationConstant += random*maxNoise;
                                                     
   // write out first inequality to file
   constraints << setprecision(10);
   constraints << -simulationConstant;
   constraints << setprecision(20);
   for (int i = 0; i < correctionCount; i++)
      if (corrections[i] != 0.0)
         {
            double coefficient = coefficients[i][location][axis];
            if (coefficient > tiny)
               {
                  constraints << " + ";
                  constraints << coefficient;
                  constraints << '*';
                  for (int j = 0; j < 3; j++)  // loop over char array
                     constraints << correctionName[i][j];
               }
            else if (coefficient < -tiny)
               {
                  constraints << " - ";
                  constraints << -coefficient;
                  constraints << '*';
                  for (int j = 0; j < 3; j++)  // loop over char array
                     constraints << correctionName[i][j];
               }
         }
   constraints << " + 1 <= " << "E;\n";

   // write out second inequality to file
   constraints << setprecision(4);
   constraints << -simulationConstant;
   constraints << setprecision(10);
   for (int i = 0; i < correctionCount; i++)
      if (corrections[i] != 0.0)
         {
            double coefficient = coefficients[i][location][axis];
            if (coefficient > tiny)
               {
                  constraints << " + ";
                  constraints << coefficient;
                  constraints << '*';
                  for (int j = 0; j < 3; j++)  // loop over char array
                     constraints << correctionName[i][j];
               }
            else if (coefficient < -tiny)
               {
                  constraints << " - ";
                  constraints << -coefficient;
                  constraints << '*';
                  for (int j = 0; j < 3; j++)  // loop over char array
                     constraints << correctionName[i][j];
               }
         }
   constraints << " - 1 >= " << "-E;\n";
}

void
HexapodMechanism::moveCutterTo(const Vector3,
                               const Vector3)
{
   // not yet implemented
}

void
HexapodMechanism::moveCutterBy(const Vector3,
                               const Vector3)
{
   // not yet implemented in C++
   // C code below based on old Pascal, converted automagically by p2c

/*
typedef struct LegType {
  double XTop, YTop, ZTop; // real coordinates
  double XBot, YBot, ZBot, Length, MaxLength, MinLength;
  int X, Y, Z, S, SquaresDifference;
} LegType;

Static double StepsPerMillimetre;

  // compute leg control variables
  Hexapod.LegPairA.LeftLeg.X = (int)
      floor((Hexapod.LegPairA.LeftLeg.XBot - Hexapod.LegPairA.LeftLeg.XTop) *
            StepsPerMillimetre + 0.5);
  Hexapod.LegPairA.LeftLeg.Y = (int)
      floor((Hexapod.LegPairA.LeftLeg.YBot - Hexapod.LegPairA.LeftLeg.YTop) *
            StepsPerMillimetre + 0.5);
  Hexapod.LegPairA.LeftLeg.Z = (int)
      floor((Hexapod.LegPairA.LeftLeg.ZBot - Hexapod.LegPairA.LeftLeg.ZTop) *
            StepsPerMillimetre + 0.5);
  TEMP = Hexapod.LegPairA.LeftLeg.XBot - Hexapod.LegPairA.LeftLeg.XTop;
  TEMP1 = Hexapod.LegPairA.LeftLeg.YBot - Hexapod.LegPairA.LeftLeg.YTop;
  TEMP2 = Hexapod.LegPairA.LeftLeg.ZBot - Hexapod.LegPairA.LeftLeg.ZTop;
  Hexapod.LegPairA.LeftLeg.Length = sqrt(
      TEMP * TEMP + TEMP1 * TEMP1 + TEMP2 * TEMP2);
  Hexapod.LegPairA.LeftLeg.S =
    (int)floor(Hexapod.LegPairA.LeftLeg.Length * StepsPerMillimetre + 0.5);
  TEMP = 1.0 * Hexapod.LegPairA.LeftLeg.S;
  TEMP1 = 1.0 * Hexapod.LegPairA.LeftLeg.X;
  TEMP2 = 1.0 * Hexapod.LegPairA.LeftLeg.Y;
  TEMP3 = 1.0 * Hexapod.LegPairA.LeftLeg.Z;
  Hexapod.LegPairA.LeftLeg.SquaresDifference = (int)floor(
        TEMP * TEMP - TEMP1 * TEMP1 - TEMP2 * TEMP2 - TEMP3 * TEMP3 + 0.5);
                              // force real calculation because of range
}


void TranslateHexapod(double dX, double dY, double dZ)
{

  // Translates the platform by the given amount. This means translating the
  // platform joints by this amount. We cheat for now and do it a direction 
  // at a time.

  // Only X axis for now
  int StepsNeeded, i;
  LegType Leg;

  StepsNeeded = (int)floor(fabs(dX) * StepsPerMillimetre + 0.5);
                                                // not proper bookkeeping
  Leg = Hexapod.LegPairA.LeftLeg;
  Count = 0; // use pointer variables for legs, or an array, for easy indexing

  if (dX >= 0)
    {// X increments
    for (i = 1; i <= StepsNeeded; i++)
      {
      Leg.SquaresDifference += -Leg.X * 2 - 1;
      Leg.X++;
      if (Leg.X >= 0)
        {// S increments
        while (Leg.SquaresDifference < 0)
          {
          Leg.SquaresDifference += Leg.S * 2 + 1;
          Leg.S++;
        }
      }
      else
        {  // X < 0, S decrements
        while (Leg.SquaresDifference > 0)
          {
          Leg.SquaresDifference += 1 - Leg.S * 2;
          Leg.S--;
        }
      }
    }
  }
  else
    {  // dX < 0, X decrements
    for (i = 1; i <= StepsNeeded; i++)
      {
      Leg.SquaresDifference += Leg.X * 2 - 1;
      Leg.X--;
      if (Leg.X >= 0)
        {// S decrements
        while (Leg.SquaresDifference > 0)
          {
          Leg.SquaresDifference += 1 - Leg.S * 2;
          Leg.S--;
        }
      }
      else
        {  // X < 0, S increments
        while (Leg.SquaresDifference < 0)
          {
          Leg.SquaresDifference += Leg.S * 2 + 1;
          Leg.S++;
        }
      }
    }
  }

  Hexapod.LegPairA.LeftLeg = Leg;

  // diagnostics
  printf("x = ");
  printf("%12d%12d", Leg.X, 6);
  printf("  s = ");
  printf("%12d%12d", Leg.S, 6);
  printf("  sqrdiff = ");
  printf("%12d%12d\n", Leg.SquaresDifference, 6);
  printf("test counter reads ");
  printf("%12d%12d\n\n", Count, 10);

}
*/

}

/*
{The main low level hexapod control algorithm, in Pascal. Not yet implemented
in C++.}

{hexapod model
    take x step}

procedure TakeXStep
  (var Leg: LegType;
   Direction: DirectionType);

{Takes a single step on the x axis in the required direction, at the foot of 
the leg. Steps the leg if appropriate.}

{The incremental integer control algorithm used is based on Pythagorus' theorem,
and is similar to Bresenham's circular interpolation algorithm. When we take a
notional step on an orthogonal axis, the algorithm decides if we must take a 
step on the leg axis and in which direction.

The algorithm is based on sqr(x + 1) = sqr(x) + 2*x + 1. That is, if you are
moving on an integer x grid, the difference between the squares on adjacent
grid lines is simply 2*x + 1, where x is the smaller value. Extending this to
an integer xy grid, the square of the length of the vector from the origin to 
the point (x, y) is sqr(x) + sqr(y). If we increment x by 1, this square changes
by 2*x + 1. This gives us a fast way of keeping track of a squared vector 
length. Similarly, considering now an integer s grid along the leg, if we take
a positive step in s, the square of the leg length changes by 2*s + 1. By
comparing these squares, we can decide whether a step in x requires one in s for
a best fit. To avoid very large integer values we use the difference of the 
squares, storing this in a variable indicated by [...] in the C++ fragment
below.

A positive step in x is required:
[sqr(s) - sqr(x) - sqr(y)] = [sqr(s) - sqr(x) - sqr(y)] + 2*x + 1;
// new value of variable        old value of variable     old x
x = x + 1;

while ([sqr(s) - sqr(x) - sqr(y)] < 0)  // max twice round loop
   {
      // take a step in s
      [sqr(s) - sqr(x) - sqr(y)] := [sqr(s) - sqr(x) - sqr(y)] + 2*s + 1;
      // new value of variable         old value of variable    old s
      s := s + 1;
      // issue hardware signal
   }

The algorithm is exact, not involving approximations. A squared form of length 
error is maintained and used for decision making. The approach is easily 
expanded to three dimensions and negative motion:

We maintain [sqr(s) - sqr(x) - sqr(y) - sqr(z)].
If x increments, subtract 2*x + 1
If x decrements, add 2*x - 1 (subtract -2*x + 1)
If s increments, add 2*s + 1
If s decrements, subtract 2*s - 1 (add -2*s + 1)
Note that x can be positive or negative, but s is always positive.}

begin
   with Leg do

   if Direction = Positive then
      if X >= 0 then {X and S increment}
         begin
            SquaresDifference := SquaresDifference - (2*X + 1);
            X := X + 1;

            while SquaresDifference < 0 do {max twice round the loop}
               begin
                  Count := Count + 1;
                 {take a + step in s, issuing hardware signal if active}
                  SquaresDifference := SquaresDifference + (2*S + 1);
                  S := S + 1 
               end;

           {diagnostics}
           {PutString('x = ');
            PutInteger(X, 6);
            PutString('  s = ');
            PutInteger(S, 6);
            NewLine;}

         end
      else {X < 0, X increments, S decrements}
         begin
            SquaresDifference := SquaresDifference - (2*X + 1);
            X := X + 1;

            while SquaresDifference > 0 do {max twice round the loop}
               begin
                  Count := Count + 1;
                 {take a - step in s, issuing hardware signal if active}
                  SquaresDifference := SquaresDifference - (2*S - 1);
                  S := S - 1 
               end;

           {diagnostics}
           {PutString('x = ');
            PutInteger(X, 6);
            PutString('  s = ');
            PutInteger(S, 6);
            NewLine;}

         end
   else {Direction = Negative}
      if X >= 0 then {X and S decrement}
         begin
            SquaresDifference := SquaresDifference + (2*X - 1);
            X := X - 1;

            while SquaresDifference > 0 do {max twice round the loop}
               begin
                  Count := Count + 1;
                 {take a - step in s, issuing hardware signal if active}
                  SquaresDifference := SquaresDifference - (2*S - 1);
                  S := S - 1 
               end;

           {diagnostics}
           {PutString('x = ');
            PutInteger(X, 6);
            PutString('  s = ');
            PutInteger(S, 6);
            NewLine;}

         end
      else {X < 0, X decrements, S increments}
         begin
            SquaresDifference := SquaresDifference + (2*X - 1);
            X := X - 1;

            while SquaresDifference < 0 do {max twice round the loop}
               begin
                  Count := Count + 1;
                 {take a + step in s, issuing hardware signal if active}
                  SquaresDifference := SquaresDifference + (2*S + 1);
                  S := S + 1 
               end;

           {diagnostics}
           {PutString('x = ');
            PutInteger(X, 6);
            PutString('  s = ');
            PutInteger(S, 6);
            NewLine;}

         end
end;
*/

