
// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

#ifndef HexapodMechanismClass
#define HexapodMechanismClass

#include <math.h>
#include <iomanip.h>
#include "Vector3.H"
#include "Vector6.H"
#include "Matrix3.H"
#include "Matrix6.H"
#include "Primitive.H"
#include "Artifact.H"
#include "HexapodLeg.H"
#include "HexapodTopJoint.H"
#include "HexapodBottomJoint.H"
#include "HexapodPlatform.H"

struct MechanismParameters
{
   double 
      topCircleRadius,  // top frame
      bottomCircleRadius,  // platform
      legPairSectorAngle,  // subtended at top centre
      supportHeight,  // top frame
      platformHeight;  // leg value
   LegParameters legParameters;
   TopJointParameters topJointParameters;
   BottomJointParameters bottomJointParameters;
   PlatformParameters platformParameters;
};

enum Selector {MIN, MAX};  // selector for data reporting functions

// calibration constants
const int correctionCount = 27;  // calibration corrections
const int locationCount = 3;  // calibration measurement locations
const int axisCount = 3;  // 3-space
const int charCount = 4;  // chars in correction names, including null

class HexapodMechanism: public Artifact
{
   public:
      HexapodMechanism() {};
      HexapodMechanism
        (char* name, 
         const MechanismParameters& mechanismParameters);

      void movePlatformTo
        (const Vector3 rotation,
         const Vector3 translation);
      void movePlatformBy
        (const Vector3 rotation, 
         const Vector3 translation);
      void moveCutterTo  // not implemented
        (const Vector3,  
         const Vector3);
      void moveCutterBy  // not implemented
        (const Vector3, 
         const Vector3);
      void applyCutterForce
        (const Vector3 force,
         const Vector3 position);

      double angleACB  // leg pair triangle bottom angle
        (Selector selector);
      double angleBAC  // leg pair triangle top angle
        (Selector selector);
      double angleGCV  // leg pair triangle to horizontal
        (Selector selector);
      double angleACF  // leg to bottom joint long axis
        (Selector selector);
      double angleGCH  // leg pair triangle to platform
        (Selector selector);
      double legForce  // axial leg force
        (Selector selector);
      double cutterDisplacement();  // absolute cutter displacement
      double resolutionError();  // absolute resolution error

      void computeCalibrationCoefficients();
      void reportRangeStatus();
      void reportAngles();
      void reportForces();
      void reportDisplacement();
      void reportResolutionError();
      void reportCorrections();
      void reportCoefficient  // calibration coefficients
        (const int correction,  
         const int location,
	 const int axis); 

      // calibration simulation predefined corrections
      void defineCorrections
        (const double values[correctionCount]);  
      // calibration simulation constraints to be solved
      void generateConstraints
        (const int location,  
	 const int axis,
	 const double maxNoise); 

   private:
      double 
         topCircleRadius, bottomCircleRadius,
         legPairSectorAngle, topHeight, bottomHeight;
      LegParameters legParameters;
      TopJointParameters topJointParameters;
      BottomJointParameters bottomJointParameters;
      PlatformParameters platformParameters;

      double baseAngle;

      Vector3 
         pointA[3], pointB[3], pointC[3],
         platformTranslation, platformRotation;

      HexapodLeg 
         leg1A, leg1B, 
         leg2A, leg2B, 
         leg3A, leg3B;

      HexapodTopJoint 
         topJoint1A, topJoint1B, 
         topJoint2A, topJoint2B, 
         topJoint3A, topJoint3B;

      HexapodBottomJoint
         bottomJoint1,
         bottomJoint2,
         bottomJoint3;

      HexapodPlatform platform;

      ArtifactInstance* leg1AInstance;
      ArtifactInstance* leg1BInstance;
      ArtifactInstance* leg2AInstance;
      ArtifactInstance* leg2BInstance;
      ArtifactInstance* leg3AInstance; 
      ArtifactInstance* leg3BInstance;
      ArtifactInstance* platformInstance;

      Vector3 cutterForce;
      Vector3 cutterForcePosition;
      double forceAC[3], forceBC[3];  // leg axial forces

      Vector3 coefficients[correctionCount][locationCount];
      double corrections[correctionCount];
      char correctionName[correctionCount][charCount];
};

#endif
