
// C++ Vector and Matrix Algebra Routines from Graphics Gems IV
// Author: Jean-Francois DOUE
// Version 3.1 --- October 1993
// Minor modifications by John Storrs

#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <ctype.h>

#include "Vector2.H"
#include "Vector3.H"
#include "Matrix3.H"

// Matrix3 Class Member Functions

// CONSTRUCTORS

Matrix3::Matrix3() {}

Matrix3::Matrix3(const Vector3& v0, const Vector3& v1, const Vector3& v2)
{ v[0] = v0; v[1] = v1; v[2] = v2; }

Matrix3::Matrix3(const double d)
{ v[0] = v[1] = v[2] = Vector3(d); }

Matrix3::Matrix3(const Matrix3& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; }


// ASSIGNMENT OPERATORS

Matrix3& Matrix3::operator = ( const Matrix3& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; return *this; }

Matrix3& Matrix3::operator += ( const Matrix3& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; return *this; }

Matrix3& Matrix3::operator -= ( const Matrix3& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; return *this; }

Matrix3& Matrix3::operator *= ( const double d )
{ v[0] *= d; v[1] *= d; v[2] *= d; return *this; }

Matrix3& Matrix3::operator /= ( const double d )
{ v[0] /= d; v[1] /= d; v[2] /= d; return *this; }

Vector3& Matrix3::operator [] ( int i) {
    if (i < VX || i > VZ)
	V_ERROR("Matrix3 [] operator: illegal access; index = " << i << '\n')
    return v[i];
}


// SPECIAL FUNCTIONS

Matrix3 Matrix3::transpose() {
    return Matrix3(Vector3(v[0][0], v[1][0], v[2][0]),
		Vector3(v[0][1], v[1][1], v[2][1]),
		Vector3(v[0][2], v[1][2], v[2][2]));
}

Matrix3 Matrix3::inverse()  // Gauss-Jordan elimination with partial pivoting
    {
    Matrix3 a(*this),  // As a evolves from original mat into identity
	 b(identity2D());  // b evolves from identity into inverse(a)
    int	 i, j, i1;

    // Loop over cols of a from left to right, eliminating above and below diag
    for (j=0; j<3; j++) {  // Find largest pivot in column j among rows j..2
    i1 = j;  // Row with largest pivot candidate
    for (i=j+1; i<3; i++)
	if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
	    i1 = i;

    // Swap rows i1 and j in a and b to put pivot on diagonal
    swap(a.v[i1], a.v[j]);
    swap(b.v[i1], b.v[j]);

    // Scale row j to have a unit diagonal
    if (a.v[j].n[j]==0.)
	V_ERROR("Matrix3::inverse: singular matrix; can't invert\n")
    b.v[j] /= a.v[j].n[j];
    a.v[j] /= a.v[j].n[j];

    // Eliminate off-diagonal elems in col j of a, doing identical ops to b
    for (i=0; i<3; i++)
	if (i!=j) {
	b.v[i] -= a.v[i].n[j]*b.v[j];
	a.v[i] -= a.v[i].n[j]*a.v[j];
	}
    }
    return b;
}

Matrix3& Matrix3::apply(V_FCT_PTR fct) {
    v[VX].apply(fct);
    v[VY].apply(fct);
    v[VZ].apply(fct);
    return *this;
}


// FRIENDS

Matrix3 operator - (const Matrix3& a)
{ return Matrix3(-a.v[0], -a.v[1], -a.v[2]); }

Matrix3 operator + (const Matrix3& a, const Matrix3& b)
{ return Matrix3(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2]); }

Matrix3 operator - (const Matrix3& a, const Matrix3& b)
{ return Matrix3(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2]); }

Matrix3 operator * (Matrix3& a, Matrix3& b) {
    #define ROWCOL(i, j) \
    a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + a.v[i].n[2]*b.v[2][j]
    return Matrix3
      (Vector3(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2)),
       Vector3(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2)),
       Vector3(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2)));
    #undef ROWCOL(i, j)
}

Matrix3 operator * (const Matrix3& a, const double d)
{ return Matrix3(a.v[0] * d, a.v[1] * d, a.v[2] * d); }

Matrix3 operator * (const double d, const Matrix3& a)
{ return a*d; }

Matrix3 operator / (const Matrix3& a, const double d)
{ return Matrix3(a.v[0] / d, a.v[1] / d, a.v[2] / d); }

int operator == (const Matrix3& a, const Matrix3& b)
{ return (a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]); }

int operator != (const Matrix3& a, const Matrix3& b)
{ return !(a == b); }

ostream& operator << (ostream& s, Matrix3& m)
{ return s << m.v[VX] << '\n' << m.v[VY] << '\n' << m.v[VZ]; }

istream& operator >> (istream& s, Matrix3& m) {
    Matrix3    m_tmp;

    s >> m_tmp[VX] >> m_tmp[VY] >> m_tmp[VZ];
    if (s)
	m = m_tmp;
    return s;
}

void swap(Matrix3& a, Matrix3& b)
{ Matrix3 tmp(a); a = b; b = tmp; }


// FUNCTIONS

Matrix3 identity2D()
{   return Matrix3
      (Vector3(1.0, 0.0, 0.0),
       Vector3(0.0, 1.0, 0.0),
       Vector3(0.0, 0.0, 1.0)); }

Matrix3 translation2D(Vector2& v)
{   return Matrix3
      (Vector3(1.0, 0.0, v[VX]),
       Vector3(0.0, 1.0, v[VY]),
       Vector3(0.0, 0.0, 1.0)); }

Matrix3 rotation2D(Vector2& Center, const double angle) 
{   double  
       c = cos(angle),
       s = sin(angle);

    return Matrix3
      (Vector3(c, -s, Center[VX] * (1.0-c) + Center[VY] * s),
       Vector3(s, c, Center[VY] * (1.0-c) - Center[VX] * s),
       Vector3(0.0, 0.0, 1.0));
}

Matrix3 scaling2D(Vector2& scaleVector)
{   return Matrix3
      (Vector3(scaleVector[VX], 0.0, 0.0),
       Vector3(0.0, scaleVector[VY], 0.0),
       Vector3(0.0, 0.0, 1.0)); }

