
// Copyright 1996 John Storrs. All rights reserved. This work may be copied, 
// modified and distributed under the terms of the GNU General Public License.

// simulateCalibration-3.C

// Z measurements on a reference plane at bed level. Readings taken at two
// levels 300mm vertically apart, using accurate probe extension pieces.
// With platform horizontal and alternately rotated PI/6 about the z axis,
// take measurements with C0, C1, C2 probes at 100mm horizontal intervals.

#include <std/ctime.h>  // for random number seeding
#include "Files.H"
#include "Matrix4.H"
#include "HexapodFrame.H"
#include "HexapodBed.H"
#include "HexapodLeg.H"
#include "HexapodTopJoint.H"
#include "HexapodBottomJoint.H"
#include "HexapodPlatform.H"
#include "HexapodMechanism.H"

void 
defineParameters(FrameParameters& frameParameters,
                 BedParameters& bedParameters,
                 MechanismParameters& mechanismParameters);

main()
{
   FrameParameters frameParameters;
   BedParameters bedParameters;
   MechanismParameters mechanismParameters;
   defineParameters(frameParameters, bedParameters, mechanismParameters);

   HexapodFrame hexapodFrame ("hexapod frame", frameParameters);
   HexapodBed hexapodBed ("hexapod bed", bedParameters);
   HexapodMechanism hexapodMechanism ("hexapod mechanism", mechanismParameters);

   // seed random number generator, for calibration noise generation
   // this should ensure different random numbers with different program
   // executions, but the effect of increasing the argument to srand should
   // be examined - does an increase of 1 make a large change to the pattern?
   int seconds = time(NULL);  // get current time 
   srand(seconds % 1000);  // cycles through 0..999

   // define max noise, mm
   double maxNoise = 0.025;

   // define the objective function for lp_solve
   constraints << "min: E;\n";

   // correction values for this simulation
   // must be all positive for now - lp_solve
   double values[correctionCount];
   values[0]  = 0.0;  // pointA[0].x  // datum
   values[1]  = 0.0;  // pointA[0].y  // datum
   values[2]  = 0.5;  // pointA[0].z
   values[3]  = 0.4;  // pointA[1].x
   values[4]  = 0.3;  // pointA[1].y
   values[5]  = 1.0;  // pointA[1].z
   values[6]  = 0.7;  // pointA[2].x
   values[7]  = 0.3;  // pointA[2].y
   values[8]  = 0.9;  // pointA[2].z
   values[9]  = 0.0;  // pointB[0].x  // datum
   values[10] = 0.3;  // pointB[0].y
   values[11] = 0.8;  // pointB[0].z
   values[12] = 0.3;  // pointB[1].x
   values[13] = 0.6;  // pointB[1].y
   values[14] = 1.3;  // pointB[1].z
   values[15] = 0.5;  // pointB[2].x
   values[16] = 0.8;  // pointB[2].y
   values[17] = 0.7;  // pointB[2].z
   values[18] = 0.4;  // lengthA[0]
   values[19] = 1.4;  // lengthA[1]
   values[20] = 0.8;  // lengthA[2]
   values[21] = 0.2;  // lengthB[0]
   values[22] = 0.5;  // lengthB[1]
   values[23] = 0.9;  // lengthB[2]
   values[24] = 0.0;  // lengthIJ (C0-C1)
   values[25] = 0.0;  // lengthJK (C1-C2)
   values[26] = 0.0;  // lengthKI (C2-C0)

   // reduce by factor of 10 if required to see effect
   //for (int i = 0; i < correctionCount; i++)
   //   values[i] /= 10;  

   // define correction values
   hexapodMechanism.defineCorrections(values);

   // report correction values
   correction << "\n\n   CALIBRATION SIMULATION CORRECTIONS - MM\n\n";
   hexapodMechanism.reportCorrections();

   for (double z = 800; z < 1200; z += 300)
      for (double y = -600; y <= 600; y += 100)
         for (double x = -600; x <= 600; x += 100)
            {
               hexapodMechanism.movePlatformTo(Vector3(0, 0, -PI/6),
	                                       Vector3(x, y, z));
	       hexapodMechanism.computeCalibrationCoefficients();
       	       hexapodMechanism.generateConstraints(0, VZ, maxNoise);
       	       hexapodMechanism.generateConstraints(1, VZ, maxNoise);
       	       hexapodMechanism.generateConstraints(2, VZ, maxNoise);

               hexapodMechanism.movePlatformTo(Vector3(0, 0, PI/6),
	                                       Vector3(x, y, z));
	       hexapodMechanism.computeCalibrationCoefficients();
       	       hexapodMechanism.generateConstraints(0, VZ, maxNoise);
       	       hexapodMechanism.generateConstraints(1, VZ, maxNoise);
       	       hexapodMechanism.generateConstraints(2, VZ, maxNoise);
	    }
}
